"""K-Means Algorithms Module."""

from __future__ import annotations

import warnings
from abc import abstractmethod
from typing import Any, Generic, Optional, Tuple, TypeVar

import numpy as np
from sklearn.base import BaseEstimator, ClusterMixin, TransformerMixin
from sklearn.utils import check_random_state
from sklearn.utils.validation import check_is_fitted

from ..._utils import RandomStateLike, _check_compatible_fdata
from ...misc.metrics import Metric, PairwiseMetric, l2_distance
from ...representation import FDataGrid
from ...representation._typing import NDArrayAny, NDArrayFloat, NDArrayInt

SelfType = TypeVar("SelfType", bound="BaseKMeans[Any]")
MembershipType = TypeVar("MembershipType", bound=NDArrayAny)


class BaseKMeans(
    BaseEstimator,  # type: ignore
    ClusterMixin,  # type: ignore
    TransformerMixin,  # type: ignore
    Generic[MembershipType],
):
    """Base class to implement K-Means clustering algorithms.

    Class from which both :class:`K-Means
    <skfda.ml.clustering.base_kmeans.KMeans>` and
    :class:`Fuzzy K-Means <skfda.ml.clustering.base_kmeans.FuzzyKMeans>`
    classes inherit.
    """

    def __init__(
        self,
        *,
        n_clusters: int = 2,
        init: Optional[FDataGrid] = None,
        metric: Metric[FDataGrid] = l2_distance,
        n_init: int = 1,
        max_iter: int = 100,
        tol: float = 1e-4,
        random_state: RandomStateLike = 0,
    ):
        """Initialize the BaseKMeans class.

        Args:
            n_clusters: Number of groups into which the samples
                are classified. Defaults to 2.
            init: Contains the initial centers of the
                different clusters the algorithm starts with. Its data_marix
                must be of the shape (n_clusters, fdatagrid.ncol,
                fdatagrid.dim_codomain). Defaults to None, and the centers are
                initialized randomly.
            metric: functional data metric. Defaults to
                *l2_distance*.
            n_init: Number of time the k-means algorithm will
                be run with different centroid seeds. The final results will
                be the best output of n_init consecutive runs in terms of
                inertia.
            max_iter: Maximum number of iterations of the
                clustering algorithm for a single run. Defaults to 100.
            tol: tolerance used to compare the centroids
                calculated with the previous ones in every single run of the
                algorithm.
            random_state:
                Determines random number generation for centroid
                initialization. Use an int to make the randomness
                deterministic. Defaults to 0.
                See :term:`Glossary <random_state>`.
        """
        self.n_clusters = n_clusters
        self.init = init
        self.metric = metric
        self.n_init = n_init
        self.max_iter = max_iter
        self.tol = tol
        self.random_state = random_state

    def _check_clustering(self, fdata: FDataGrid) -> FDataGrid:
        """Check the arguments used in fit.

        Args:
            fdata: Object whose samples
                are classified into different groups.

        Returns:
            Validated input.

        """
        if fdata.dim_domain > 1:
            raise NotImplementedError(
                "Only support 1 dimension on the domain.",
            )

        if fdata.n_samples < 2:
            raise ValueError(
                "The number of observations must be greater than 1.",
            )

        if self.n_clusters < 2:
            raise ValueError(
                "The number of clusters must be greater than 1.",
            )

        if self.n_init < 1:
            raise ValueError(
                "The number of iterations must be greater than 0.",
            )

        if self.init is not None and self.n_init != 1:
            self.n_init = 1
            warnings.warn(
                "Warning: The number of iterations is ignored "
                "because the init parameter is set.",
            )

        if (
            self.init is not None
            and self.init.data_matrix.shape != (
                (self.n_clusters,) + fdata.data_matrix.shape[1:]
            )
        ):
            raise ValueError(
                "The init FDataGrid data_matrix should be of "
                "shape (n_clusters, n_features, dim_codomain) "
                "and gives the initial centers.",
            )

        if self.max_iter < 1:
            raise ValueError(
                "The number of maximum iterations must be greater than 0.",
            )

        if self.tol < 0:
            raise ValueError("The tolerance must be positive.")

        return fdata

    def _tolerance(self, fdata: FDataGrid) -> float:
        variance = fdata.var()
        mean_variance = np.mean(variance[0].data_matrix)

        return mean_variance * self.tol

    def _init_centroids(
        self,
        fdatagrid: FDataGrid,
        random_state: np.random.RandomState,
    ) -> FDataGrid:
        """
        Compute the initial centroids.

        Args:
            fdatagrid: Object whose samples are classified into different
                groups.
            random_state: Random number generation for centroid initialization.

        Returns:
            Initial centroids.

        """
        if self.init is None:
            _, idx = np.unique(
                fdatagrid.data_matrix,
                axis=0,
                return_index=True,
            )
            unique_data = fdatagrid[np.sort(idx)]

            if len(unique_data) < self.n_clusters:
                raise ValueError(
                    "Not enough unique data points to "
                    "initialize the requested number of "
                    "clusters",
                )

            indices = random_state.permutation(len(unique_data))[
                :self.n_clusters
            ]
            centroids = unique_data[indices]

            return centroids.copy()

        return self.init.copy()

    def _check_params(self) -> None:
        pass

    @abstractmethod
    def _create_membership(self, n_samples: int) -> MembershipType:
        pass

    @abstractmethod
    def _update(
        self,
        fdata: FDataGrid,
        membership_matrix: MembershipType,
        distances_to_centroids: NDArrayFloat,
        centroids: FDataGrid,
    ) -> None:
        pass

    def _algorithm(
        self,
        fdata: FDataGrid,
        random_state: np.random.RandomState,
    ) -> Tuple[NDArrayFloat, FDataGrid, NDArrayFloat, int]:
        """
        Fuzzy K-Means algorithm.

        Implementation of the Fuzzy K-Means algorithm for FDataGrid objects
        of any dimension.

        Args:
            fdata: Object whose samples are clustered,
                classified into different groups.
            random_state: random number generation for
                centroid initialization.

        Returns:
            Tuple containing:

                membership values:
                    membership value that observation has to each cluster.

                centroids:
                    centroids for each cluster.

                distances_to_centroids: distances of each sample to each
                    cluster.

                repetitions: number of iterations the algorithm was run.

        """
        repetitions = 0
        centroids_old_matrix = np.zeros(
            (self.n_clusters,) + fdata.data_matrix.shape[1:],
        )
        membership_matrix = self._create_membership(fdata.n_samples)

        centroids = self._init_centroids(fdata, random_state)
        centroids_old = centroids.copy(data_matrix=centroids_old_matrix)

        pairwise_metric = PairwiseMetric(self.metric)

        tolerance = self._tolerance(fdata)

        while (
            repetitions == 0
            or (
                not np.all(self.metric(centroids, centroids_old) < tolerance)
                and repetitions < self.max_iter
            )
        ):

            centroids_old.data_matrix[...] = centroids.data_matrix

            distances_to_centroids = pairwise_metric(fdata, centroids)

            self._update(
                fdata=fdata,
                membership_matrix=membership_matrix,
                distances_to_centroids=distances_to_centroids,
                centroids=centroids,
            )

            repetitions += 1

        return (
            membership_matrix,
            centroids,
            distances_to_centroids,
            repetitions,
        )

    @abstractmethod
    def _compute_inertia(
        self,
        membership: MembershipType,
        centroids: FDataGrid,
        distances_to_centroids: NDArrayFloat,
    ) -> float:
        pass

    def fit(
        self: SelfType,
        X: FDataGrid,
        y: None = None,
        sample_weight: None = None,
    ) -> SelfType:
        """
        Fit the model.

        Args:
            X: Object whose samples are clusered,
                classified into different groups.
            y: present here for API consistency by convention.
            sample_weight: present here for API consistency by
                convention.

        Returns:
            Fitted model.

        """
        fdata = self._check_clustering(X)
        random_state = check_random_state(self.random_state)

        self._check_params()

        best_inertia = np.inf

        for _ in range(self.n_init):
            (
                membership,
                centroids,
                distances_to_centroids,
                n_iter,
            ) = (
                self._algorithm(
                    fdata=fdata,
                    random_state=random_state,
                )
            )

            inertia = self._compute_inertia(
                membership,
                centroids,
                distances_to_centroids,
            )

            if inertia < best_inertia:
                best_inertia = inertia
                best_membership = membership
                best_centroids = centroids
                best_distances_to_centroids = distances_to_centroids
                best_n_iter = n_iter

        self._best_membership = best_membership
        self.labels_ = self._prediction_from_membership(best_membership)
        self.cluster_centers_ = best_centroids
        self._distances_to_centers = best_distances_to_centroids
        self.inertia_ = best_inertia
        self.n_iter_ = best_n_iter

        return self

    def _predict_membership(
        self,
        X: FDataGrid,
        sample_weight: None = None,
    ) -> MembershipType:
        """Predict the closest cluster each sample in X belongs to.

        Args:
            X: Object whose samples are classified into different groups.
            sample_weight: present here for API consistency by convention.

        Returns:
            Label of each sample.

        """
        check_is_fitted(self)
        _check_compatible_fdata(self.cluster_centers_, X)

        membership_matrix = self._create_membership(X.n_samples)
        centroids = self.cluster_centers_.copy()

        pairwise_metric = PairwiseMetric(self.metric)

        distances_to_centroids = pairwise_metric(X, centroids)

        self._update(
            fdata=X,
            membership_matrix=membership_matrix,
            distances_to_centroids=distances_to_centroids,
            centroids=centroids,
        )

        return membership_matrix

    @abstractmethod
    def _prediction_from_membership(
        self,
        membership_matrix: MembershipType,
    ) -> NDArrayInt:
        pass

    def predict(
        self,
        X: FDataGrid,
        sample_weight: None = None,
    ) -> NDArrayInt:
        """Predict the closest cluster each sample in X belongs to.

        Args:
            X: Object whose samples are classified into different groups.
            sample_weight: present here for API consistency by convention.

        Returns:
            Label of each sample.

        """
        return self._prediction_from_membership(
            self._predict_membership(X, sample_weight),
        )

    def transform(self, X: FDataGrid) -> NDArrayFloat:
        """Transform X to a cluster-distance space.

        Args:
            X: Object whose samples are classified into
                different groups.

        Returns:
            distances_to_centers:
                distances of each sample to each cluster.

        """
        check_is_fitted(self)
        _check_compatible_fdata(self.cluster_centers_, X)
        return self._distances_to_centers

    def fit_transform(
        self,
        X: FDataGrid,
        y: None = None,
        sample_weight: None = None,
    ) -> NDArrayFloat:
        """Compute clustering and transform X to cluster-distance space.

        Args:
            X: Object whose samples are classified into different groups.
            y: present here for API consistency by convention.
            sample_weight: present here for API consistency by convention.

        Returns:
            Distances of each sample to each cluster.

        """
        self.fit(X)
        return self._distances_to_centers

    def score(
        self,
        X: FDataGrid,
        y: None = None,
        sample_weight: None = None,
    ) -> float:
        """Opposite of the value of X on the K-means objective.

        Args:
            X: Object whose samples are classified into
                different groups.
            y: present here for API consistency by convention.
            sample_weight: present here for API consistency by
                convention.

        Returns:
            Negative ``inertia_`` attribute.

        """
        check_is_fitted(self)
        _check_compatible_fdata(self.cluster_centers_, X)
        return -self.inertia_


class KMeans(BaseKMeans[NDArrayInt]):
    r"""K-Means algorithm for functional data.

    Let :math:`\mathbf{X = \left\{ x_{1}, x_{2}, ..., x_{n}\right\}}` be a
    given dataset to be analyzed, and :math:`\mathbf{V = \left\{ v_{1}, v_{2},
    ..., v_{c}\right\}}` be the set of centers of clusters in
    :math:`\mathbf{X}` dataset in :math:`m` dimensional space :math:`\left(
    \mathbb{R}^m \right)`. Where :math:`n` is the number of objects, :math:`m`
    is the number of features, and :math:`c` is the number of partitions or
    clusters.

    KM iteratively computes cluster centroids in order to minimize the sum with
    respect to the specified measure. KM algorithm aims at minimizing an
    objective function known as the squared error function given as follows:

    .. math::
        J_{KM}\left(\mathbf{X}; \mathbf{V}\right) = \sum_{i=1}^{c}
        \sum_{j=1}^{n}D_{ij}^2

    Where, :math:`D_{ij}^2` is the squared chosen distance measure which can
    be any p-norm: :math:`D_{ij} = \lVert x_{ij} - v_{i} \rVert = \left(
    \int_I \lvert x_{ij} - v_{i}\rvert^p dx \right)^{ \frac{1}{p}}`, being
    :math:`I` the domain where :math:`\mathbf{X}` is defined, :math:`1
    \leqslant i \leqslant c`, :math:`1 \leqslant j\leqslant n_{i}`. Where
    :math:`n_{i}` represents the number of data points in i-th cluster.

    For :math:`c` clusters, KM is based on an iterative algorithm minimizing
    the sum of distances from each observation to its cluster centroid. The
    observations are moved between clusters until the sum cannot be decreased
    any more. KM algorithm involves the following steps:

    1. Centroids of :math:`c` clusters are chosen from :math:`\mathbf{X}`
        randomly or are passed to the function as a parameter.

    2. Distances between data points and cluster centroids are calculated.

    3. Each data point is assigned to the cluster whose centroid is
        closest to it.

    4. Cluster centroids are updated by using the following formula:
        :math:`\mathbf{v_{i}} ={\sum_{i=1}^{n_{i}}x_{ij}}/n_{i}` :math:`1
        \leqslant i \leqslant c`.

    5. Distances from the updated cluster centroids are recalculated.

    6. If no data point is assigned to a new cluster the run of algorithm is
        stopped, otherwise the steps from 3 to 5 are repeated for probable
        movements of data points between the clusters.

    This algorithm is applied for each dimension on the image of the FDataGrid
    object.

    Args:
        n_clusters: Number of groups into which the samples are
            classified. Defaults to 2.
        init: Contains the initial centers of the
            different clusters the algorithm starts with. Its data_marix must
            be of the shape (n_clusters, fdatagrid.ncol,
            fdatagrid.dim_codomain). Defaults to None, and the centers are
            initialized randomly.
        metric: functional data metric. Defaults to
            *l2_distance*.
        n_init: Number of time the k-means algorithm will be
            run with different centroid seeds. The final results will be the
            best output of n_init consecutive runs in terms of inertia.
        max_iter: Maximum number of iterations of the
            clustering algorithm for a single run. Defaults to 100.
        tol: Tolerance used to compare the centroids
            calculated with the previous ones in every single run of the
            algorithm.
        random_state:
            Determines random number generation for centroid initialization.
            Use an int to make the randomness deterministic. Defaults to 0.
            See :term:`Glossary <random_state>`.

    Attributes:
        labels\_: Vector in which each entry contains the cluster each
            observation belongs to.
        cluster_centers\_: data_matrix of shape (n_clusters, ncol,
            dim_codomain) and contains the centroids for each cluster.
        inertia\_: Sum of squared distances of samples to their closest
            cluster center for each dimension.
        n_iter\_: number of iterations the algorithm was run for each
            dimension.

    Example:

        >>> import skfda
        >>> data_matrix = [[1, 1, 2, 3, 2.5, 2],
        ...                [0.5, 0.5, 1, 2, 1.5, 1],
        ...                [-1, -1, -0.5, 1, 1, 0.5],
        ...                [-0.5, -0.5, -0.5, -1, -1, -1]]
        >>> grid_points = [0, 2, 4, 6, 8, 10]
        >>> fd = skfda.FDataGrid(data_matrix, grid_points)
        >>> kmeans = skfda.ml.clustering.KMeans(random_state=0)
        >>> kmeans.fit(fd)
        KMeans(...)
        >>> kmeans.cluster_centers_.data_matrix
        array([[[ 0.16666667],
                [ 0.16666667],
                [ 0.83333333],
                [ 2.        ],
                [ 1.66666667],
                [ 1.16666667]],
               [[-0.5       ],
                [-0.5       ],
                [-0.5       ],
                [-1.        ],
                [-1.        ],
                [-1.        ]]])

    """

    def _compute_inertia(
        self,
        membership: NDArrayInt,
        centroids: FDataGrid,
        distances_to_centroids: NDArrayFloat,
    ) -> float:
        distances_to_their_center = np.choose(
            membership,
            distances_to_centroids.T,
        )

        return np.sum(distances_to_their_center**2)

    def _create_membership(self, n_samples: int) -> NDArrayInt:
        return np.empty(n_samples, dtype=int)

    def _prediction_from_membership(
        self,
        membership_matrix: NDArrayInt,
    ) -> NDArrayInt:
        return membership_matrix

    def _update(
        self,
        fdata: FDataGrid,
        membership_matrix: NDArrayInt,
        distances_to_centroids: NDArrayFloat,
        centroids: FDataGrid,
    ) -> None:

        membership_matrix[:] = np.argmin(distances_to_centroids, axis=1)

        for i in range(self.n_clusters):

            indices = np.where(membership_matrix == i)[0]

            if len(indices) != 0:
                centroids.data_matrix[i] = np.average(
                    fdata.data_matrix[indices, ...],
                    axis=0,
                )


class FuzzyCMeans(BaseKMeans[NDArrayFloat]):
    r"""
    Fuzzy c-Means clustering for functional data.

    Let :math:`\mathbf{X = \left\{ x_{1}, x_{2}, ..., x_{n}\right\}}` be a
    given dataset to be analyzed, and :math:`\mathbf{V = \left\{ v_{1}, v_{2},
    ..., v_{c}\right\}}` be the set of centers of clusters in
    :math:`\mathbf{X}` dataset in :math:`m` dimensional space :math:`\left(
    \mathbb{R}^m \right)`. Where :math:`n` is the number of objects, :math:`m`
    is the number of features, and :math:`c` is the number of partitions
    or clusters.

    FCM minimizes the following objective function:

    .. math::
        J_{FCM}\left(\mathbf{X}; \mathbf{U, V}\right) = \sum_{i=1}^{c}
        \sum_{j=1}^{n}u_{ij}^{f}D_{ij}^2.

    This function differs from classical KM with the use of weighted squared
    errors instead of using squared errors only. In the objective function,
    :math:`\mathbf{U}` is a fuzzy partition matrix that is computed from
    dataset :math:`\mathbf{X}`: :math:`\mathbf{U} = [u_{ij}] \in M_{FCM}`.

    The fuzzy clustering of :math:`\mathbf{X}` is represented with
    :math:`\mathbf{U}` membership matrix. The element :math:`u_{ij}` is the
    membership value of j-th object to i-th cluster. In this case, the i-th row
    of :math:`\mathbf{U}` matrix is formed with membership values of :math:`n`
    objects to i-th cluster. :math:`\mathbf{V}` is a prototype vector of
    cluster prototypes (centroids): :math:`\mathbf{V = \left\{ v_{1}, v_{2},
    ..., v_{c}\right\}}`,:math:`\mathbf{v_{i}}\in \mathbb{R}^m`.

    :math:`D_{ij}^2` is the squared chosen distance measure which can be any
    p-norm: :math:`D_{ij} =\lVert x_{ij} - v_{i} \rVert = \left( \int_I \lvert
    x_{ij} - v_{i}\rvert^p dx \right)^{ \frac{1}{p}}`, being :math:`I` the
    domain where :math:`\mathbf{X}` is defined, :math:`1 \leqslant i
    \leqslant c`, :math:`1 \leqslant j\leqslant n_{i}`. Where :math:`n_{i}`
    represents the number of data points in i-th cluster.

    FCM is an iterative process and stops when the number of iterations is
    reached to maximum, or when the centroids of the clusters do not change.
    The steps involved in FCM are:

        1. Centroids of :math:`c` clusters are chosen from :math:`\mathbf{X}`
            randomly or are passed to the function as a parameter.

        2. Membership values of data points to each cluster are calculated
            with: :math:`u_{ij} = \left[ \sum_{k=1}^c\left( D_{ij}/D_{kj}
            \right)^\frac{2}{f-1} \right]^{-1}`.

        3. Cluster centroids are updated by using the following formula:
            :math:`\mathbf{v_{i}} =\frac{\sum_{j=1}^{n}u_{ij}^f x_{j}}{
            \sum_{j=1}^{n} u_{ij}^f}`, :math:`1 \leqslant i \leqslant c`.

        4. If no cluster centroid changes the run of algorithm is stopped,
            otherwise return to step 2.

    This algorithm is applied for each dimension on the image of the FDataGrid
    object.

    Args:
        n_clusters: Number of groups into which the samples are
            classified. Defaults to 2.
        init: Contains the initial centers of the
            different clusters the algorithm starts with. Its data_marix must
            be of the shape (n_clusters, fdatagrid.ncol,
            fdatagrid.dim_codomain). Defaults to None, and the centers are
            initialized randomly.
        metric: functional data metric. Defaults to
            *l2_distance*.
        n_init: Number of time the k-means algorithm will be
            run with different centroid seeds. The final results will be the
            best output of n_init consecutive runs in terms of inertia.
        max_iter: Maximum number of iterations of the
            clustering algorithm for a single run. Defaults to 100.
        tol: tolerance used to compare the centroids
            calculated with the previous ones in every single run of the
            algorithm.
        random_state:
            Determines random number generation for centroid initialization.
            Use an int to make the randomness deterministic. Defaults to 0.
            See :term:`Glossary <random_state>`.
        fuzzifier: Scalar parameter used to specify the
            degree of fuzziness in the fuzzy algorithm. Defaults to 2.

    Attributes:
        membership_degree\_: Matrix in which each entry contains the
            probability of belonging to each group.
        labels\_: Vector in which each entry contains the cluster each
            observation belongs to (the one with the maximum membership
            degree).
        cluster_centers\_: data_matrix of shape
            (n_clusters, ncol, dim_codomain) and contains the centroids for
            each cluster.
        inertia\_: Sum of squared
            distances of samples to their closest cluster center for each
            dimension.
        n_iter\_: number of iterations
            the algorithm was run for each dimension.


    Example:

        >>> import skfda
        >>> data_matrix = [[[1, 0.3], [2, 0.4], [3, 0.5], [4, 0.6]],
        ...                [[2, 0.5], [3, 0.6], [4, 0.7], [5, 0.7]],
        ...                [[3, 0.2], [4, 0.3], [5, 0.4], [6, 0.5]]]
        >>> grid_points = [2, 4, 6, 8]
        >>> fd = skfda.FDataGrid(data_matrix, grid_points)
        >>> fuzzy_kmeans = skfda.ml.clustering.FuzzyCMeans(random_state=0)
        >>> fuzzy_kmeans.fit(fd)
        FuzzyCMeans(...)
        >>> fuzzy_kmeans.cluster_centers_.data_matrix
        array([[[ 2.83994301,  0.24786354],
                [ 3.83994301,  0.34786354],
                [ 4.83994301,  0.44786354],
                [ 5.83994301,  0.53191927]],
               [[ 1.25134384,  0.35023779],
                [ 2.25134384,  0.45023779],
                [ 3.25134384,  0.55023779],
                [ 4.25134384,  0.6251158 ]]])


    """

    def __init__(
        self,
        *,
        n_clusters: int = 2,
        init: Optional[FDataGrid] = None,
        metric: Metric[FDataGrid] = l2_distance,
        n_init: int = 1,
        max_iter: int = 100,
        tol: float = 1e-4,
        random_state: RandomStateLike = 0,
        fuzzifier: float = 2,
    ) -> None:
        super().__init__(
            n_clusters=n_clusters,
            init=init,
            metric=metric,
            n_init=n_init,
            max_iter=max_iter,
            tol=tol,
            random_state=random_state,
        )

        self.fuzzifier = fuzzifier

    @property
    def membership_degree_(self) -> NDArrayFloat:
        return self._best_membership

    def _check_params(self) -> None:
        if self.fuzzifier <= 1:
            raise ValueError("The fuzzifier parameter must be greater than 1.")

    def _compute_inertia(
        self,
        membership: NDArrayFloat,
        centroids: FDataGrid,
        distances_to_centroids: NDArrayFloat,
    ) -> float:
        return np.sum(
            membership**self.fuzzifier * distances_to_centroids**2,
        )

    def _create_membership(self, n_samples: int) -> NDArrayFloat:
        return np.empty((n_samples, self.n_clusters))

    def _prediction_from_membership(
        self,
        membership_matrix: NDArrayFloat,
    ) -> NDArrayInt:
        return np.argmax(membership_matrix, axis=1)

    def _update(
        self,
        fdata: FDataGrid,
        membership_matrix: NDArrayFloat,
        distances_to_centroids: NDArrayFloat,
        centroids: FDataGrid,
    ) -> None:
        # Divisions by zero allowed
        with np.errstate(divide='ignore'):
            distances_to_centers_raised = (
                distances_to_centroids**(2 / (1 - self.fuzzifier))
            )

        # Divisions infinity by infinity allowed
        with np.errstate(invalid='ignore'):
            membership_matrix[:, :] = (
                distances_to_centers_raised
                / np.sum(
                    distances_to_centers_raised,
                    axis=1,
                    keepdims=True,
                )
            )

        # inf / inf divisions should be 1 in this context
        membership_matrix[np.isnan(membership_matrix)] = 1

        membership_matrix_raised = np.power(
            membership_matrix,
            self.fuzzifier,
        )

        slice_denominator = (
            (slice(None),) + (np.newaxis,) * (fdata.data_matrix.ndim - 1)
        )
        centroids.data_matrix[:] = (
            np.einsum(
                'ij,i...->j...',
                membership_matrix_raised,
                fdata.data_matrix,
            )
            / np.sum(membership_matrix_raised, axis=0)[slice_denominator]
        )

    def predict_proba(
        self,
        X: FDataGrid,
        sample_weight: None = None,
    ) -> NDArrayFloat:
        """Predict the probability of belonging to each cluster.

        Args:
            X: Object whose samples are classified into different groups.
            sample_weight: present here for API consistency by convention.

        Returns:
            Probability of belonging to each cluster for each sample.

        """
        return self._predict_membership(X, sample_weight)



from __future__ import annotations

from typing import Tuple, Union, overload

import numpy as np
from sklearn.utils import check_random_state
from typing_extensions import Literal

from ... import concatenate
from ..._utils import RandomStateLike
from ...datasets import make_gaussian_process
from ...misc.metrics import lp_distance
from ...representation import FData, FDataGrid
from ...representation._typing import ArrayLike


def v_sample_stat(fd: FData, weights: ArrayLike, p: int = 2) -> float:
    r"""
    Compute sample statistic.

    Calculates a statistic that measures the variability between groups of
    samples in a :class:`skfda.representation.FData` object.

    The statistic defined as below is calculated between all the samples in a
    :class:`skfda.representation.FData` object with a given set of
    weights.

    Let :math:`\{f_i\}_{i=1}^k` be a set of samples in a FData object.
    Let :math:`\{w_j\}_{j=1}^k` be a set of weights, where :math:`w_i` is
    related to the sample :math:`f_i` for :math:`i=1,\dots,k`.
    The statistic is defined as:

    .. math::
        V_n = \sum_{i<j}^kw_i\|f_i-f_j\|^2

    This statistic is defined in Cuevas :footcite:`cuevas++_2004_anova`.

    Args:
         fd: Object containing all the samples for which we want
            to calculate the statistic.
         weights: Weights related to each sample. Each
            weight is expected to appear in the same position as its
            corresponding sample in the FData object.
         p: p of the lp norm. Must be greater or equal
            than 1. If p='inf' or p=np.inf it is used the L infinity metric.
            Defaults to 2.

    Returns:
        The value of the statistic.

    Raises:
        ValueError

    Examples:
        >>> from skfda.inference.anova import v_sample_stat
        >>> from skfda.representation.grid import FDataGrid
        >>> import numpy as np

        We create different trajectories to be applied in the statistic and a
        set of weights.

        >>> t = np.linspace(0, 1, 50)
        >>> x1 = t * (1 - t) ** 5
        >>> x2 = t ** 2 * (1 - t) ** 4
        >>> x3 = t ** 3 * (1 - t) ** 3
        >>> fd = FDataGrid([x1, x2, x3], grid_points=t)
        >>> weights = [10, 20, 30]

        Finally the value of the statistic is calculated:

        >>> v_sample_stat(fd, weights)
        0.01649448843348894

    References:
        .. footbibliography::

    """
    weights = np.asarray(weights)
    if not isinstance(fd, FData):
        raise ValueError("Argument type must inherit FData.")
    if len(weights) != fd.n_samples:
        raise ValueError("Number of weights must match number of samples.")

    t_ind = np.tril_indices(fd.n_samples, -1)
    coef = weights[t_ind[1]]
    return float(np.sum(
        coef * lp_distance(
            fd[t_ind[0]],
            fd[t_ind[1]],
            p=p,
        ) ** p,
    ))


def v_asymptotic_stat(fd: FData, weights: ArrayLike, p: int = 2) -> float:
    r"""
    Compute asymptitic statistic.

    Calculates a statistic that measures the variability between groups of
    samples in a :class:`skfda.representation.FData` object.

    The statistic defined as below is calculated between all the samples in a
    :class:`skfda.representation.FData` object with a given set of
    weights.

    Let :math:`\{f_i\}_{i=1}^k` be a set of samples in a FData object.
    Let :math:`\{w_j\}_{j=1}^k` be a set of weights, where :math:`w_i` is
    related to the sample :math:`f_i` for :math:`i=1,\dots,k`.
    The statistic is defined as:

    .. math::
        \sum_{i<j}^k\|f_i-f_j\sqrt{\cfrac{w_i}{w_j}}\|^2

    This statistic is defined in Cuevas :footcite:`cuevas++_2004_anova`.

    Args:
         fd: Object containing all the samples for which we want
            to calculate the statistic.
         weights: Weights related to each sample. Each
            weight is expected to appear in the same position as its
            corresponding sample in the FData object.
         p: p of the lp norm. Must be greater or equal
            than 1. If p='inf' or p=np.inf it is used the L infinity metric.
            Defaults to 2.

    Returns:
        The value of the statistic.

    Raises:
        ValueError

    Examples:
        >>> from skfda.inference.anova import v_asymptotic_stat
        >>> from skfda.representation.grid import FDataGrid
        >>> import numpy as np

        We create different trajectories to be applied in the statistic and a
        set of weights.

        >>> t = np.linspace(0, 1, 50)
        >>> x1 = t * (1 - t) ** 5
        >>> x2 = t ** 2 * (1 - t) ** 4
        >>> x3 = t ** 3 * (1 - t) ** 3
        >>> fd = FDataGrid([x1, x2, x3], grid_points=t)
        >>> weights = [10, 20, 30]

        Finally the value of the statistic is calculated:

        >>> v_asymptotic_stat(fd, weights)
        0.0018159320335885969

    References:
        .. footbibliography::

    """
    weights = np.asarray(weights)
    if not isinstance(fd, FData):
        raise ValueError("Argument type must inherit FData.")
    if len(weights) != fd.n_samples:
        raise ValueError("Number of weights must match number of samples.")
    if np.count_nonzero(weights) != len(weights):
        raise ValueError("All weights must be non-zero.")

    t_ind = np.tril_indices(fd.n_samples, -1)
    coef = np.sqrt(weights[t_ind[1]] / weights[t_ind[0]])
    left_fd = fd[t_ind[1]]
    right_fd = fd[t_ind[0]] * coef
    return float(np.sum(lp_distance(left_fd, right_fd, p=p) ** p))


def _anova_bootstrap(
    fd_grouped: Tuple[FData, ...],
    n_reps: int,
    random_state: RandomStateLike = None,
    p: int = 2,
    equal_var: bool = True,
) -> np.ndarray:

    n_groups = len(fd_grouped)
    if n_groups < 2:
        raise ValueError("At least two groups must be passed in fd_grouped.")

    for fd in fd_grouped[1:]:
        if not np.array_equal(fd.domain_range, fd_grouped[0].domain_range):
            raise ValueError(
                "Domain range must match for every FData in fd_grouped.",
            )

    start, stop = fd_grouped[0].domain_range[0]

    # List with sizes of each group
    sizes = [fd.n_samples for fd in fd_grouped]

    # Instance a random state object in case random_state is an int
    random_state = check_random_state(random_state)

    if equal_var:
        k_est = concatenate(fd_grouped).cov().data_matrix[0, ..., 0]
        k_est = [k_est] * len(fd_grouped)
    else:
        # Estimating covariances for each group
        k_est = [fd.cov().data_matrix[0, ..., 0] for fd in fd_grouped]

    # Number of sample points for gaussian processes have to match
    # the features of the covariances.
    n_features = k_est[0].shape[0]

    # Simulating n_reps observations for each of the n_groups gaussian
    # processes
    sim = [
        make_gaussian_process(
            n_reps,
            n_features=n_features,
            start=start,
            stop=stop,
            cov=k_est[i],
            random_state=random_state,
        )
        for i in range(n_groups)
    ]

    v_samples = np.empty(n_reps)
    for i in range(n_reps):
        fd = FDataGrid([s.data_matrix[i, ..., 0] for s in sim])
        v_samples[i] = v_asymptotic_stat(fd, sizes, p=p)
    return v_samples


@overload
def oneway_anova(
    *args: FData,
    n_reps: int = 2000,
    return_dist: Literal[False] = False,
    random_state: RandomStateLike = None,
    p: int = 2,
    equal_var: bool = True,
) -> Tuple[float, float]:
    pass


@overload
def oneway_anova(
    *args: FData,
    n_reps: int = 2000,
    return_dist: Literal[True],
    random_state: RandomStateLike = None,
    p: int = 2,
    equal_var: bool = True,
) -> Tuple[float, float, np.ndarray]:
    pass


def oneway_anova(
    *args: FData,
    n_reps: int = 2000,
    return_dist: bool = False,
    random_state: RandomStateLike = None,
    p: int = 2,
    equal_var: bool = True,
) -> Union[Tuple[float, float], Tuple[float, float, np.ndarray]]:
    r"""
    Perform one-way functional ANOVA.

    This function implements an asymptotic method to test the following
    null hypothesis:

    Let :math:`\{X_i\}_{i=1}^k` be a set of :math:`k` independent samples
    each one with :math:`n_i` trajectories, and let :math:`E(X_i) = m_i(
    t)`. The null hypothesis is defined as:

    .. math::
        H_0: m_1(t) = \dots = m_k(t)

    This function calculates the value of the statistic
    :func:`~skfda.inference.anova.v_sample_stat` :math:`V_n` with the means
    of the given samples. Under the null hypothesis this statistic is
    asymptotically equivalent to
    :func:`~skfda.inference.anova.v_asymptotic_stat`, where each sample
    is replaced by a gaussian process, with mean zero and the same
    covariance function as the original.

    The simulation of the distribution of the asymptotic statistic :math:`V` is
    implemented using a bootstrap procedure. One observation of the
    :math:`k` different gaussian processes defined above is simulated,
    and the value of :func:`~skfda.inference.anova.v_asymptotic_stat` is
    calculated. This procedure is repeated `n_reps` times, creating a
    sampling distribution of the statistic.

    This procedure is from Cuevas :footcite:`cuevas++_2004_anova`.

    Args:
        args: The sample measurements for each each group.
        n_reps: Number of simulations for the bootstrap
            procedure. Defaults to 2000 (This value may change in future
            versions).
        return_dist: Flag to indicate if the function should
            return a numpy.array with the sampling distribution simulated.
        random_state: Random state.
        p: p of the lp norm. Must be greater or equal
            than 1. If p='inf' or p=np.inf it is used the L infinity metric.
            Defaults to 2.
        equal_var: If True (default), perform a One-way
            ANOVA assuming the same covariance operator for all the groups,
            else considers an independent covariance operator for each group.

    Returns:
        Tuple containing the value of the sample statistic, p-value (and
        sampling distribution of the simulated asymptotic statistic if
        `return_dist` is `True`).

    Examples:
        >>> from skfda.inference.anova import oneway_anova
        >>> from skfda.datasets import fetch_gait
        >>> from numpy.random import RandomState
        >>> from numpy import printoptions

        >>> fd = fetch_gait()["data"].coordinates[1]
        >>> fd1, fd2, fd3 = fd[:13], fd[13:26], fd[26:]
        >>> oneway_anova(fd1, fd2, fd3, random_state=RandomState(42))
        (179.52499999999998, 0.5945)
        >>> _, _, dist = oneway_anova(fd1, fd2, fd3, n_reps=3,
        ...     random_state=RandomState(42),
        ...     return_dist=True)
        >>> with printoptions(precision=4):
        ...     print(dist)
        [ 184.0698 212.7395  195.3663]

    References:
        .. footbibliography::

    """
    if len(args) < 2:
        raise ValueError("At least two groups must be passed as parameter.")
    if not all(isinstance(fd, FData) for fd in args):
        raise ValueError("Argument type must inherit FData.")
    if n_reps < 1:
        raise ValueError("Number of simulations must be positive.")

    fd_groups = args
    if not all(isinstance(fd, type(fd_groups[0])) for fd in fd_groups[1:]):
        raise TypeError('Found mixed FData types in arguments.')

    for fd in fd_groups[1:]:
        if not np.array_equal(fd.domain_range, fd_groups[0].domain_range):
            raise ValueError("Domain range must match for every FData passed.")

    if isinstance(fd_groups[0], FDataGrid):
        # Creating list with all the sample points
        list_sample = [fd.grid_points[0].tolist() for fd in fd_groups]
        # Checking that the all the entries in the list are the same
        if list_sample.count(list_sample[0]) != len(list_sample):
            raise ValueError(
                "All FDataGrid passed must have the same grid points.",
            )
    else:  # If type is FDataBasis, check same basis
        list_basis = [fd.basis for fd in fd_groups]
        if list_basis.count(list_basis[0]) != len(list_basis):
            raise NotImplementedError(
                "Not implemented for FDataBasis with different basis.",
            )

    # FData where each sample is the mean of each group
    fd_means = concatenate([fd.mean() for fd in fd_groups])

    # Base statistic
    vn = v_sample_stat(fd_means, [fd.n_samples for fd in fd_groups], p=p)

    # Computing sampling distribution
    simulation = _anova_bootstrap(
        fd_groups,
        n_reps,
        random_state=random_state,
        p=p,
        equal_var=equal_var,
    )

    p_value = np.sum(simulation > vn) / len(simulation)

    if return_dist:
        return vn, p_value, simulation

    return vn, p_value



import itertools
from typing import Any, Iterable, Tuple

import numpy as np

from ._basis import Basis


class Tensor(Basis):
    r"""Tensor basis.

    Basis for multivariate functions constructed as a tensor product of
    :math:`\mathbb{R} \to \mathbb{R}` bases.


    Attributes:
        domain_range (tuple): a tuple of length ``dim_domain`` containing
            the range of input values for each dimension.
        n_basis (int): number of functions in the basis.

    Examples:
        Defines a tensor basis over the interval :math:`[0, 5] \times [0, 3]`
        consisting on the functions

        .. math::

            1, v, u, uv, u^2, u^2v

        >>> from skfda.representation.basis import Tensor, Monomial
        >>>
        >>> basis_x = Monomial(domain_range=(0,5), n_basis=3)
        >>> basis_y = Monomial(domain_range=(0,3), n_basis=2)
        >>>
        >>> basis = Tensor([basis_x, basis_y])


        And evaluates all the functions in the basis in a list of descrete
        values.

        >>> basis([(0., 2.), (3., 0), (2., 3.)])
        array([[[  1.],
                [  1.],
                [  1.]],
               [[  2.],
                [  0.],
                [  3.]],
               [[  0.],
                [  3.],
                [  2.]],
               [[  0.],
                [  0.],
                [  6.]],
               [[  0.],
                [  9.],
                [  4.]],
               [[  0.],
                [  0.],
                [ 12.]]])

    """

    def __init__(self, basis_list: Iterable[Basis]):

        self._basis_list = tuple(basis_list)

        if not all(
            b.dim_domain == 1 and b.dim_codomain == 1
            for b in self._basis_list
        ):
            raise ValueError(
                "The basis functions must be univariate and scalar valued",
            )

        super().__init__(
            domain_range=[b.domain_range[0] for b in basis_list],
            n_basis=np.prod([b.n_basis for b in basis_list]),
        )

    @property
    def basis_list(self) -> Tuple[Basis, ...]:
        return self._basis_list

    def _evaluate(self, eval_points: np.ndarray) -> np.ndarray:

        matrix = np.zeros((self.n_basis, len(eval_points), self.dim_codomain))

        basis_evaluations = [
            b(eval_points[:, i:i + 1])
            for i, b in enumerate(self.basis_list)
        ]

        for i, ev in enumerate(itertools.product(*basis_evaluations)):

            matrix[i, :, :] = np.prod(ev, axis=0)

        return matrix

    def _gram_matrix(self) -> np.ndarray:

        gram_matrices = [b.gram_matrix() for b in self.basis_list]

        gram = gram_matrices[0]

        for g in gram_matrices[1:]:
            n_rows = len(gram) * len(g)
            gram = np.multiply.outer(gram, g)
            gram = np.moveaxis(gram, [1, 2], [2, 1])
            gram = gram.reshape(n_rows, n_rows)

        return gram

    def __eq__(self, other: Any) -> bool:
        return super().__eq__(other) and self.basis_list == other.basis_list

    def __hash__(self) -> int:
        return hash((super().__hash__(), self.basis_list))



from skfda import FDataGrid, concatenate
from skfda.exploratory import stats
import unittest

from mpl_toolkits.mplot3d import axes3d
import scipy.stats.mstats

import numpy as np


class TestFDataGrid(unittest.TestCase):

    # def setUp(self): could be defined for set up before any test

    def test_init(self):
        fd = FDataGrid([[1, 2, 3, 4, 5], [2, 3, 4, 5, 6]])
        np.testing.assert_array_equal(
            fd.data_matrix[..., 0],
            np.array([[1, 2, 3, 4, 5], [2, 3, 4, 5, 6]]))
        np.testing.assert_array_equal(fd.sample_range, [(0, 1)])
        np.testing.assert_array_equal(
            fd.grid_points, np.array([[0., 0.25, 0.5, 0.75, 1.]]))

    def test_copy_equals(self):
        fd = FDataGrid([[1, 2, 3, 4, 5], [2, 3, 4, 5, 6]])
        self.assertTrue(fd.equals(fd.copy()))

    def test_mean(self):
        fd = FDataGrid([[1, 2, 3, 4, 5], [2, 3, 4, 5, 6]])
        mean = stats.mean(fd)
        np.testing.assert_array_equal(
            mean.data_matrix[0, ..., 0],
            np.array([1.5, 2.5, 3.5, 4.5, 5.5]))
        np.testing.assert_array_equal(fd.sample_range, [(0, 1)])
        np.testing.assert_array_equal(
            fd.grid_points,
            np.array([[0., 0.25, 0.5, 0.75, 1.]]))

    def test_gmean(self):
        fd = FDataGrid([[1, 2, 3, 4, 5], [2, 3, 4, 5, 6]])
        mean = stats.gmean(fd)
        np.testing.assert_array_equal(
            mean.data_matrix[0, ..., 0],
            scipy.stats.mstats.gmean(
                np.array([[1, 2, 3, 4, 5], [2, 3, 4, 5, 6]])))
        np.testing.assert_array_equal(fd.sample_range, [(0, 1)])
        np.testing.assert_array_equal(
            fd.grid_points,
            np.array([[0., 0.25, 0.5, 0.75, 1.]]))

    def test_slice(self):
        t = (5, 3)
        fd = FDataGrid(data_matrix=np.ones(t))
        fd = fd[1:3]
        np.testing.assert_array_equal(
            fd.data_matrix[..., 0],
            np.array([[1, 1, 1], [1, 1, 1]]))

    def test_concatenate(self):
        fd1 = FDataGrid([[1, 2, 3, 4, 5], [2, 3, 4, 5, 6]])
        fd2 = FDataGrid([[3, 4, 5, 6, 7], [4, 5, 6, 7, 8]])

        fd1.argument_names = ["x"]
        fd1.coordinate_names = ["y"]
        fd = fd1.concatenate(fd2)

        np.testing.assert_equal(fd.n_samples, 4)
        np.testing.assert_equal(fd.dim_codomain, 1)
        np.testing.assert_equal(fd.dim_domain, 1)
        np.testing.assert_array_equal(fd.data_matrix[..., 0],
                                      [[1, 2, 3, 4, 5], [2, 3, 4, 5, 6],
                                       [3, 4, 5, 6, 7], [4, 5, 6, 7, 8]])
        np.testing.assert_array_equal(fd1.argument_names, fd.argument_names)
        np.testing.assert_array_equal(
            fd1.coordinate_names, fd.coordinate_names)

    def test_concatenate_coordinates(self):
        fd1 = FDataGrid([[1, 2, 3, 4], [2, 3, 4, 5]])
        fd2 = FDataGrid([[3, 4, 5, 6], [4, 5, 6, 7]])

        fd1.argument_names = ["x"]
        fd1.coordinate_names = ["y"]
        fd2.argument_names = ["w"]
        fd2.coordinate_names = ["t"]
        fd = fd1.concatenate(fd2, as_coordinates=True)

        np.testing.assert_equal(fd.n_samples, 2)
        np.testing.assert_equal(fd.dim_codomain, 2)
        np.testing.assert_equal(fd.dim_domain, 1)

        np.testing.assert_array_equal(fd.data_matrix,
                                      [[[1, 3], [2, 4], [3, 5], [4, 6]],
                                       [[2, 4], [3, 5], [4, 6], [5, 7]]])

        # Testing labels
        np.testing.assert_array_equal(("y", "t"), fd.coordinate_names)
        fd2.coordinate_names = None
        fd = fd1.concatenate(fd2, as_coordinates=True)
        np.testing.assert_array_equal(("y", None), fd.coordinate_names)
        fd1.coordinate_names = None
        fd = fd1.concatenate(fd2, as_coordinates=True)
        np.testing.assert_equal((None, None), fd.coordinate_names)

    def test_concatenate2(self):
        sample1 = np.arange(0, 10)
        sample2 = np.arange(10, 20)
        fd1 = FDataGrid([sample1])
        fd2 = FDataGrid([sample2])

        fd1.argument_names = ["x"]
        fd1.coordinate_names = ["y"]
        fd = concatenate([fd1, fd2])

        np.testing.assert_equal(fd.n_samples, 2)
        np.testing.assert_equal(fd.dim_codomain, 1)
        np.testing.assert_equal(fd.dim_domain, 1)
        np.testing.assert_array_equal(fd.data_matrix[..., 0], [sample1,
                                                               sample2])
        np.testing.assert_array_equal(fd1.argument_names, fd.argument_names)
        np.testing.assert_array_equal(
            fd1.coordinate_names, fd.coordinate_names)

    def test_coordinates(self):
        fd1 = FDataGrid([[1, 2, 3, 4], [2, 3, 4, 5]])
        fd1.argument_names = ["x"]
        fd1.coordinate_names = ["y"]
        fd2 = FDataGrid([[3, 4, 5, 6], [4, 5, 6, 7]])
        fd = fd1.concatenate(fd2, as_coordinates=True)

        # Indexing with number
        np.testing.assert_array_equal(fd.coordinates[0].data_matrix,
                                      fd1.data_matrix)
        np.testing.assert_array_equal(fd.coordinates[1].data_matrix,
                                      fd2.data_matrix)

        # Iteration
        for fd_j, fd_i in zip([fd1, fd2], fd.coordinates):
            np.testing.assert_array_equal(fd_j.data_matrix, fd_i.data_matrix)

        fd3 = fd1.concatenate(fd2, fd1, fd, as_coordinates=True)

        # Multiple indexation
        np.testing.assert_equal(fd3.dim_codomain, 5)
        np.testing.assert_array_equal(fd3.coordinates[:2].data_matrix,
                                      fd.data_matrix)
        np.testing.assert_array_equal(fd3.coordinates[-2:].data_matrix,
                                      fd.data_matrix)
        np.testing.assert_array_equal(
            fd3.coordinates[np.array(
                (False, False, True, False, True))].data_matrix,
            fd.data_matrix)

    def test_add(self):
        fd1 = FDataGrid([[1, 2, 3, 4], [2, 3, 4, 5]])

        fd2 = fd1 + fd1
        np.testing.assert_array_equal(fd2.data_matrix[..., 0],
                                      [[2, 4, 6, 8], [4, 6, 8, 10]])

        fd2 = fd1 + 2
        np.testing.assert_array_equal(fd2.data_matrix[..., 0],
                                      [[3, 4, 5, 6], [4, 5, 6, 7]])

        fd2 = fd1 + np.array(2)
        np.testing.assert_array_equal(fd2.data_matrix[..., 0],
                                      [[3, 4, 5, 6], [4, 5, 6, 7]])

        fd2 = fd1 + np.array([2])
        np.testing.assert_array_equal(fd2.data_matrix[..., 0],
                                      [[3, 4, 5, 6], [4, 5, 6, 7]])

        fd2 = fd1 + np.array([1, 2])
        np.testing.assert_array_equal(fd2.data_matrix[..., 0],
                                      [[2, 3, 4, 5], [4, 5, 6, 7]])

    def test_composition(self):
        X, Y, Z = axes3d.get_test_data(1.2)

        data_matrix = [Z.T]
        grid_points = [X[0, :], Y[:, 0]]

        g = FDataGrid(data_matrix, grid_points)
        self.assertEqual(g.dim_domain, 2)
        self.assertEqual(g.dim_codomain, 1)

        t = np.linspace(0, 2 * np.pi, 100)

        data_matrix = [10 * np.array([np.cos(t), np.sin(t)]).T]
        f = FDataGrid(data_matrix, t)
        self.assertEqual(f.dim_domain, 1)
        self.assertEqual(f.dim_codomain, 2)

        gof = g.compose(f)
        self.assertEqual(gof.dim_domain, 1)
        self.assertEqual(gof.dim_codomain, 1)


class TestEvaluateFDataGrid(unittest.TestCase):

    def setUp(self):
        data_matrix = np.array(
            [
                [
                    [[0, 1, 2], [0, 1, 2]],
                    [[0, 1, 2], [0, 1, 2]]
                ],
                [
                    [[3, 4, 5], [3, 4, 5]],
                    [[3, 4, 5], [3, 4, 5]]
                ]
            ])

        grid_points = [[0, 1], [0, 1]]

        fd = FDataGrid(data_matrix, grid_points=grid_points)
        self.assertEqual(fd.n_samples, 2)
        self.assertEqual(fd.dim_domain, 2)
        self.assertEqual(fd.dim_codomain, 3)

        self.fd = fd

    def test_evaluate_aligned(self):

        res = self.fd([(0, 0), (1, 1), (2, 2), (3, 3)])
        expected = np.array([[[0, 1, 2], [0, 1, 2], [0, 1, 2], [0, 1, 2]],
                             [[3, 4, 5], [3, 4, 5], [3, 4, 5], [3, 4, 5]]])

        np.testing.assert_allclose(res, expected)

    def test_evaluate_unaligned(self):

        res = self.fd([[(0, 0), (1, 1), (2, 2), (3, 3)],
                       [(1, 7), (5, 2), (3, 4), (6, 1)]],
                      aligned=False)
        expected = np.array([[[0, 1, 2], [0, 1, 2], [0, 1, 2], [0, 1, 2]],
                             [[3, 4, 5], [3, 4, 5], [3, 4, 5], [3, 4, 5]]])

        np.testing.assert_allclose(res, expected)

    def test_evaluate_unaligned_ragged(self):

        res = self.fd([[(0, 0), (1, 1), (2, 2), (3, 3)],
                       [(1, 7), (5, 2), (3, 4)]],
                      aligned=False)
        expected = ([[[0, 1, 2], [0, 1, 2], [0, 1, 2], [0, 1, 2]],
                     [[3, 4, 5], [3, 4, 5], [3, 4, 5]]])

        self.assertEqual(len(res), self.fd.n_samples)

        for r, e in zip(res, expected):
            np.testing.assert_allclose(r, e)

    def test_evaluate_grid_aligned(self):

        res = self.fd([[0, 1], [1, 2]], grid=True)
        expected = np.array([[[[0, 1, 2], [0, 1, 2]], [[0, 1, 2], [0, 1, 2]]],
                             [[[3, 4, 5], [3, 4, 5]], [[3, 4, 5], [3, 4, 5]]]])

        np.testing.assert_allclose(res, expected)

    def test_evaluate_grid_unaligned(self):

        res = self.fd([[[0, 1], [1, 2]], [[3, 4], [5, 6]]],
                      grid=True, aligned=False)
        expected = np.array([[[[0, 1, 2], [0, 1, 2]], [[0, 1, 2], [0, 1, 2]]],
                             [[[3, 4, 5], [3, 4, 5]], [[3, 4, 5], [3, 4, 5]]]])

        np.testing.assert_allclose(res, expected)

    def test_evaluate_grid_unaligned_ragged(self):

        res = self.fd([[[0, 1], [1, 2]], [[3, 4], [5]]],
                      grid=True, aligned=False)
        expected = ([[[[0, 1, 2], [0, 1, 2]], [[0, 1, 2], [0, 1, 2]]],
                     [[[3, 4, 5]], [[3, 4, 5]]]])

        for r, e in zip(res, expected):
            np.testing.assert_allclose(r, e)


if __name__ == '__main__':
    print()
    unittest.main()



"""Defines the most commonly used kernels."""
import math
from scipy import stats
import numpy as np


__author__ = "Miguel Carbajo Berrocal"
__email__ = "miguel.carbajo@estudiante.uam.es"


def normal(u):
    r"""Evaluate a normal kernel.

    .. math::
        K(x) = \frac{1}{\sqrt{2\pi}}e^{-\frac{x^2}{2}}

    """
    return stats.norm.pdf(u)


def cosine(u):
    r"""Cosine kernel.

    .. math::
        K(x) =
        \begin{cases}
        \frac{\pi}{4} cos\left( \frac{\pi x}{2} \right) & \mbox{if } |x| \le
        1 \\
        0 & \mbox{elsewhere}
        \end{cases}

    """
    if isinstance(u, np.ndarray):
        res = np.zeros(u.shape)
        res[abs(u) <= 1] = math.pi / 4 * (math.cos(math.pi * u[abs(u) <= 1]
                                                   / 2))
        return res
    if abs(u) <= 1:
        return math.pi / 4 * (math.cos(math.pi * u / 2))
    return 0


def epanechnikov(u):
    r"""Epanechnikov kernel.

    .. math::
        K(x) =
        \begin{cases}
        0.75(1-x^2) & \mbox{if } |x| \le 1 \\
        0 & \mbox{elsewhere}
        \end{cases}

    """
    if isinstance(u, np.ndarray):
        res = np.zeros(u.shape)
        res[abs(u) <= 1] = 0.75 * (1 - u[abs(u) <= 1] ** 2)
        return res
    if abs(u) <= 1:
        return 0.75 * (1 - u ** 2)
    return 0


def tri_weight(u):
    r"""Tri-weight kernel.

    .. math::
        K(x) =
        \begin{cases}
        \frac{35}{32} \left(1 - u^2 \right) ^3 & \mbox{if } |x| \le 1 \\
        0 & \mbox{elsewhere}
        \end{cases}

    """
    if isinstance(u, np.ndarray):
        res = np.zeros(u.shape)
        res[abs(u) <= 1] = 35 / 32 * (1 - u[abs(u) <= 1] ** 2) ** 3
        return res
    if abs(u) <= 1:
        return 35 / 32 * (1 - u ** 2) ** 3
    return 0


def quartic(u):
    r"""Quartic kernel.

    .. math::
        K(x) =
        \begin{cases}
        \frac{15}{16} \left( 1- u^2 \right) ^2 & \mbox{if } |x| \le 1 \\
        0 & \mbox{elsewhere}
        \end{cases}

    """
    if isinstance(u, np.ndarray):
        res = np.zeros(u.shape)
        res[abs(u) <= 1] = 15 / 16 * (1 - u[abs(u) <= 1] ** 2) ** 2
        return res
    if abs(u) <= 1:
        return 15 / 16 * (1 - u ** 2) ** 2
    return 0


def uniform(u):
    r"""Uniform kernel.

    .. math::
        K(x) =
        \begin{cases}
        0.5 & \mbox{if } |x| \le
        1 \\
        0 & \mbox{elsewhere}
        \end{cases}

    """
    if isinstance(u, np.ndarray):
        res = np.zeros(u.shape)
        res[abs(u) <= 1] = 0.5
        return res
    if abs(u) <= 1:
        return 0.5
    return 0



"""Test neighbors classifiers and regressors."""

import unittest

import numpy as np

from skfda.datasets import make_multimodal_samples, make_sinusoidal_process
from skfda.exploratory.outliers import LocalOutlierFactor  # Pending theory
from skfda.exploratory.stats import mean
from skfda.misc.metrics import PairwiseMetric, l2_distance
from skfda.ml.classification import (
    KNeighborsClassifier,
    NearestCentroid,
    RadiusNeighborsClassifier,
)
from skfda.ml.clustering import NearestNeighbors
from skfda.ml.regression import KNeighborsRegressor, RadiusNeighborsRegressor
from skfda.representation.basis import Fourier


class TestNeighbors(unittest.TestCase):

    def setUp(self):
        """Create test data."""
        random_state = np.random.RandomState(0)
        modes_location = np.concatenate((
            random_state.normal(-0.3, 0.04, size=15),
            random_state.normal(0.3, 0.04, size=15),
        ))

        idx = np.arange(30)
        random_state.shuffle(idx)

        modes_location = modes_location[idx]
        self.modes_location = modes_location
        self.y = np.array(15 * [0] + 15 * [1])[idx]

        self.X = make_multimodal_samples(
            n_samples=30,
            modes_location=modes_location,
            noise=0.05,
            random_state=random_state,
        )
        self.X2 = make_multimodal_samples(
            n_samples=30,
            modes_location=modes_location,
            noise=0.05,
            random_state=1,
        )

        self.probs = np.array(15 * [[1.0, 0.0]] + 15 * [[0.0, 1.0]])[idx]

        # Dataset with outliers
        fd_clean = make_sinusoidal_process(
            n_samples=25,
            error_std=0,
            phase_std=0.1,
            random_state=0,
        )
        fd_outliers = make_sinusoidal_process(
            n_samples=2,
            error_std=0,
            phase_mean=0.5,
            random_state=5,
        )
        self.fd_lof = fd_outliers.concatenate(fd_clean)

    def test_predict_classifier(self):
        """Tests predict for neighbors classifier."""
        for neigh in (
            KNeighborsClassifier(),
            RadiusNeighborsClassifier(radius=0.1),
            NearestCentroid(),
            NearestCentroid(metric=l2_distance, centroid=mean),
        ):

            neigh.fit(self.X, self.y)
            pred = neigh.predict(self.X)
            np.testing.assert_array_equal(
                pred,
                self.y,
                err_msg=f'fail in {type(neigh)}',
            )

    def test_predict_proba_classifier(self):
        """Tests predict proba for k neighbors classifier."""
        neigh = KNeighborsClassifier(metric=l2_distance)

        neigh.fit(self.X, self.y)
        probs = neigh.predict_proba(self.X)

        np.testing.assert_array_almost_equal(probs, self.probs)

    def test_predict_regressor(self):
        """Test scalar regression, predicts mode location."""
        # Dummy test, with weight = distance, only the sample with distance 0
        # will be returned, obtaining the exact location
        knnr = KNeighborsRegressor(weights='distance')
        rnnr = RadiusNeighborsRegressor(weights='distance', radius=0.1)

        knnr.fit(self.X, self.modes_location)
        rnnr.fit(self.X, self.modes_location)

        np.testing.assert_array_almost_equal(
            knnr.predict(self.X),
            self.modes_location,
        )
        np.testing.assert_array_almost_equal(
            rnnr.predict(self.X),
            self.modes_location,
        )

    def test_kneighbors(self):
        """Test k neighbor searches for all k-neighbors estimators."""
        nn = NearestNeighbors()
        nn.fit(self.X)

        lof = LocalOutlierFactor(n_neighbors=5)
        lof.fit(self.X)

        knn = KNeighborsClassifier()
        knn.fit(self.X, self.y)

        knnr = KNeighborsRegressor()
        knnr.fit(self.X, self.modes_location)

        for neigh in (nn, knn, knnr, lof):

            dist, links = neigh.kneighbors(self.X[:4])

            np.testing.assert_array_equal(
                links,
                [[0, 7, 21, 23, 15],
                 [1, 12, 19, 18, 17],
                 [2, 17, 22, 27, 26],
                 [3, 4, 9, 5, 25],
                 ],
            )

            graph = neigh.kneighbors_graph(self.X[:4])

            dist_kneigh = l2_distance(self.X[0], self.X[7])

            np.testing.assert_array_almost_equal(dist[0, 1], dist_kneigh)

            for i in range(30):
                self.assertEqual(graph[0, i] == 1.0, i in links[0])
                self.assertEqual(graph[0, i] == 0.0, i not in links[0])

    def test_radius_neighbors(self):
        """Test query with radius."""
        nn = NearestNeighbors(radius=0.1)
        nn.fit(self.X)

        knn = RadiusNeighborsClassifier(radius=0.1)
        knn.fit(self.X, self.y)

        knnr = RadiusNeighborsRegressor(radius=0.1)
        knnr.fit(self.X, self.modes_location)

        for neigh in (nn, knn, knnr):

            dist, links = neigh.radius_neighbors(self.X[:4])

            np.testing.assert_array_equal(links[0], np.array([0, 7]))
            np.testing.assert_array_equal(links[1], np.array([1]))
            np.testing.assert_array_equal(links[2], np.array([2, 17, 22, 27]))
            np.testing.assert_array_equal(links[3], np.array([3, 4, 9]))

            dist_kneigh = l2_distance(self.X[0], self.X[7])

            np.testing.assert_array_almost_equal(dist[0][1], dist_kneigh)

            graph = neigh.radius_neighbors_graph(self.X[:4])

            for i in range(30):
                self.assertEqual(graph[0, i] == 1.0, i in links[0])
                self.assertEqual(graph[0, i] == 0.0, i not in links[0])

    def test_knn_functional_response(self):
        knnr = KNeighborsRegressor(n_neighbors=1)

        knnr.fit(self.X, self.X)

        res = knnr.predict(self.X)
        np.testing.assert_array_almost_equal(
            res.data_matrix,
            self.X.data_matrix,
        )

    def test_knn_functional_response_sklearn(self):
        # Check sklearn metric
        knnr = KNeighborsRegressor(
            n_neighbors=1,
            metric='euclidean',
            multivariate_metric=True,
        )
        knnr.fit(self.X, self.X)

        res = knnr.predict(self.X)
        np.testing.assert_array_almost_equal(
            res.data_matrix,
            self.X.data_matrix,
        )

    def test_knn_functional_response_precomputed(self):
        knnr = KNeighborsRegressor(
            n_neighbors=4,
            weights='distance',
            metric='precomputed',
        )
        d = PairwiseMetric(l2_distance)
        distances = d(self.X[:4], self.X[:4])

        knnr.fit(distances, self.X[:4])

        res = knnr.predict(distances)
        np.testing.assert_array_almost_equal(
            res.data_matrix, self.X[:4].data_matrix,
        )

    def test_radius_functional_response(self):
        knnr = RadiusNeighborsRegressor(
            metric=l2_distance,
            weights='distance',
        )

        knnr.fit(self.X, self.X)

        res = knnr.predict(self.X)
        np.testing.assert_array_almost_equal(
            res.data_matrix, self.X.data_matrix,
        )

    def test_functional_response_custom_weights(self):

        knnr = KNeighborsRegressor(weights=self._weights, n_neighbors=5)
        response = self.X.to_basis(Fourier(domain_range=(-1, 1), n_basis=10))
        knnr.fit(self.X, response)

        res = knnr.predict(self.X)
        np.testing.assert_array_almost_equal(
            res.coefficients, response.coefficients,
        )

    def test_functional_regression_distance_weights(self):

        knnr = KNeighborsRegressor(
            weights='distance', n_neighbors=10,
        )
        knnr.fit(self.X[:10], self.X[:10])
        res = knnr.predict(self.X[11])

        d = PairwiseMetric(l2_distance)
        distances = d(self.X[:10], self.X[11]).flatten()

        weights = 1 / distances
        weights /= weights.sum()

        response = (self.X[:10] * weights).sum()
        np.testing.assert_array_almost_equal(
            res.data_matrix, response.data_matrix,
        )

    def test_functional_response_basis(self):
        knnr = KNeighborsRegressor(weights='distance', n_neighbors=5)
        response = self.X.to_basis(Fourier(domain_range=(-1, 1), n_basis=10))
        knnr.fit(self.X, response)

        res = knnr.predict(self.X)
        np.testing.assert_array_almost_equal(
            res.coefficients, response.coefficients,
        )

    def test_radius_outlier_functional_response(self):
        knnr = RadiusNeighborsRegressor(radius=0.001)
        knnr.fit(self.X[3:6], self.X[3:6])

        # No value given
        with np.testing.assert_raises(ValueError):
            knnr.predict(self.X[:10])

        # Test response
        knnr = RadiusNeighborsRegressor(
            radius=0.001, outlier_response=self.X[0],
        )
        knnr.fit(self.X[:6], self.X[:6])

        res = knnr.predict(self.X[:7])
        np.testing.assert_array_almost_equal(
            self.X[0].data_matrix, res[6].data_matrix,
        )

    def test_nearest_centroids_exceptions(self):

        # Test more than one class
        nn = NearestCentroid()
        with np.testing.assert_raises(ValueError):
            nn.fit(self.X[:3], 3 * [0])

        # Precomputed not supported
        nn = NearestCentroid(metric='precomputed')
        with np.testing.assert_raises(ValueError):
            nn.fit(self.X[:3], 3 * [0])

    def test_functional_regressor_exceptions(self):

        knnr = RadiusNeighborsRegressor()

        with np.testing.assert_raises(ValueError):
            knnr.fit(self.X[:3], self.X[:4])

    def test_search_neighbors_precomputed(self):
        d = PairwiseMetric(l2_distance)
        distances = d(self.X[:4], self.X[:4])

        nn = NearestNeighbors(metric='precomputed', n_neighbors=2)
        nn.fit(distances, self.y[:4])

        _, neighbors = nn.kneighbors(distances)

        np.testing.assert_array_almost_equal(
            neighbors,
            np.array([[0, 3], [1, 2], [2, 1], [3, 0]]),
        )

    def test_search_neighbors_sklearn(self):

        nn = NearestNeighbors(
            metric='euclidean',
            multivariate_metric=True,
            n_neighbors=2,
        )
        nn.fit(self.X[:4], self.y[:4])

        _, neighbors = nn.kneighbors(self.X[:4])

        np.testing.assert_array_almost_equal(
            neighbors,
            np.array([[0, 3], [1, 2], [2, 1], [3, 0]]),
        )

    def test_score_scalar_response(self):

        neigh = KNeighborsRegressor()

        neigh.fit(self.X, self.modes_location)
        r = neigh.score(self.X, self.modes_location)
        np.testing.assert_almost_equal(r, 0.9975889963743335)

    def test_score_functional_response(self):

        neigh = KNeighborsRegressor()

        y = 5 * self.X + 1
        neigh.fit(self.X, y)
        r = neigh.score(self.X, y)
        np.testing.assert_almost_equal(r, 0.962651178452408)

        # Weighted case and basis form
        y = y.to_basis(Fourier(domain_range=y.domain_range[0], n_basis=5))
        neigh.fit(self.X, y)

        r = neigh.score(
            self.X[:7],
            y[:7],
            sample_weight=4 * [1.0 / 5] + 3 * [1.0 / 15],
        )
        np.testing.assert_almost_equal(r, 0.9982527586114364)

    def test_score_functional_response_exceptions(self):
        neigh = RadiusNeighborsRegressor()
        neigh.fit(self.X, self.X)

        with np.testing.assert_raises(ValueError):
            neigh.score(self.X, self.X, sample_weight=[1, 2, 3])

    def test_multivariate_response_score(self):

        neigh = RadiusNeighborsRegressor()
        y = make_multimodal_samples(n_samples=5, dim_domain=2, random_state=0)
        neigh.fit(self.X[:5], y)

        # It is not supported the multivariate score by the moment
        with np.testing.assert_raises(ValueError):
            neigh.score(self.X[:5], y)

    def test_lof_fit_predict(self):
        """Test same results with different forms to call fit_predict."""
        # Outliers
        expected = np.ones(len(self.fd_lof))
        expected[:2] = -1

        # With default l2 distance
        lof = LocalOutlierFactor()
        res = lof.fit_predict(self.fd_lof)
        np.testing.assert_array_equal(expected, res)

        # With explicit l2 distance
        lof2 = LocalOutlierFactor(metric=l2_distance)
        res2 = lof2.fit_predict(self.fd_lof)
        np.testing.assert_array_equal(expected, res2)

        d = PairwiseMetric(l2_distance)
        distances = d(self.fd_lof, self.fd_lof)

        # With precompute distances
        lof3 = LocalOutlierFactor(metric='precomputed')
        res3 = lof3.fit_predict(distances)
        np.testing.assert_array_equal(expected, res3)

        # With multivariate sklearn
        lof4 = LocalOutlierFactor(metric='euclidean', multivariate_metric=True)
        res4 = lof4.fit_predict(self.fd_lof)
        np.testing.assert_array_equal(expected, res4)

        # Other way of call fit_predict, undocumented in sklearn
        lof5 = LocalOutlierFactor(novelty=True)
        res5 = lof5.fit(self.fd_lof).predict()
        np.testing.assert_array_equal(expected, res5)

        # Check values of negative outlier factor
        negative_lof = [  # noqa: WPS317
            -7.1068, -1.5412, -0.9961,
            -0.9854, -0.9896, -1.0993,
            -1.065, -0.9871, -0.9821,
            -0.9955, -1.0385, -1.0072,
            -0.9832, -1.0134, -0.9939,
            -1.0074, -0.992, -0.992,
            -0.9883, -1.0012, -1.1149,
            -1.002, -0.9994, -0.9869,
            -0.9726, -0.9989, -0.9904,
        ]

        np.testing.assert_array_almost_equal(
            lof.negative_outlier_factor_.round(4), negative_lof,
        )

        # Check same negative outlier factor
        np.testing.assert_array_almost_equal(
            lof.negative_outlier_factor_,
            lof2.negative_outlier_factor_,
        )

        np.testing.assert_array_almost_equal(
            lof.negative_outlier_factor_,
            lof3.negative_outlier_factor_,
        )

    def test_lof_decision_function(self):
        """Test decision function and score samples of LOF."""
        lof = LocalOutlierFactor(novelty=True)
        lof.fit(self.fd_lof[5:])

        score = lof.score_samples(self.fd_lof[:5])

        np.testing.assert_array_almost_equal(
            score.round(4),
            [-5.9726, -1.3445, -0.9853, -0.9817, -0.985],
            err_msg='Error in LocalOutlierFactor.score_samples',
        )

        # Test decision_function = score_function - offset
        np.testing.assert_array_almost_equal(
            lof.decision_function(self.fd_lof[:5]),
            score - lof.offset_,
            err_msg='Error in LocalOutlierFactor.decision_function',
        )

    def test_lof_exceptions(self):
        """Test error due to novelty attribute."""
        lof = LocalOutlierFactor(novelty=True)

        # Error in fit_predict function
        with np.testing.assert_raises(AttributeError):
            lof.fit_predict(self.fd_lof[5:])

        lof.set_params(novelty=False)
        lof.fit(self.fd_lof[5:])

        # Error in predict function
        with np.testing.assert_raises(AttributeError):
            lof.predict(self.fd_lof[5:])

    def _weights(self, weights_):
        return np.array([w == 0 for w in weights_], dtype=float)


if __name__ == '__main__':
    unittest.main()



from typing import Optional, Sequence, Union

from matplotlib.axes import Axes
from matplotlib.figure import Figure

from skfda.exploratory.visualization.representation import GraphPlot
from skfda.representation import FData

from ._baseplot import BasePlot


class FPCAPlot(BasePlot):
    """
    FPCAPlot visualization.

    Args:
        mean: The functional data object containing the mean function.
            If len(mean) > 1, the mean is computed.
        components: The principal components
        multiple: Multiple of the principal component curve to be added or
            subtracted.
        fig: Figure over which the graph is plotted. If not specified it will
            be initialized
        axes: Axes over where the graph is  plotted.
            If ``None``, see param fig.
        n_rows: Designates the number of rows of the figure.
        n_cols: Designates the number of columns of the figure.
    """

    def __init__(
        self,
        mean: FData,
        components: FData,
        multiple: float,
        chart: Union[Figure, Axes, None] = None,
        *,
        fig: Optional[Figure] = None,
        axes: Optional[Axes] = None,
        n_rows: Optional[int] = None,
        n_cols: Optional[int] = None,
    ):
        super().__init__(
            chart,
            fig=fig,
            axes=axes,
            n_rows=n_rows,
            n_cols=n_cols,
        )
        self.mean = mean
        self.components = components
        self.multiple = multiple

    @property
    def n_subplots(self) -> int:
        return len(self.components)

    def _plot(
        self,
        fig: Figure,
        axes: Sequence[Axes],
    ) -> None:

        if len(self.mean) > 1:
            self.mean = self.mean.mean()

        for i, ax in enumerate(axes):
            perturbations = self._get_component_perturbations(i)
            GraphPlot(fdata=perturbations, axes=ax).plot()
            ax.set_title(f"Principal component {i + 1}")

    def _get_component_perturbations(self, index: int = 0) -> FData:
        """
        Compute the perturbations over the mean of a principal component.

        Args:
            index: Index of the component for which we want to compute the
                perturbations

        Returns:
            The mean function followed by the positive perturbation and
            the negative perturbation.
        """
        if not isinstance(self.mean, FData):
            raise AttributeError("X must be a FData object")
        perturbations = self.mean.copy()
        perturbations = perturbations.concatenate(
            perturbations[0] + self.multiple * self.components[index],
        )
        return perturbations.concatenate(
            perturbations[0] - self.multiple * self.components[index],
        )



# -*- coding: utf-8 -*-
"""Basis smoother.

This module contains the class for the basis smoothing.

"""
from __future__ import annotations

from typing import Optional

import numpy as np
from typing_extensions import Final

from ..._utils import _cartesian_product, _to_grid_points
from ...misc.lstsq import LstsqMethod, solve_regularized_weighted_lstsq
from ...misc.regularization import L2Regularization
from ...representation import FData, FDataBasis, FDataGrid
from ...representation._typing import GridPointsLike, NDArrayFloat
from ...representation.basis import Basis
from ._linear import _LinearSmoother


class BasisSmoother(_LinearSmoother):
    r"""
    Transform raw data to a smooth functional form.

    Takes functional data in a discrete form and makes an approximates it
    to the closest function that can be generated by the basis.a.

    The fit is made so as to reduce the penalized sum of squared errors
    [RS05-5-2-6]_:

    .. math::

        PENSSE(c) = (y - \Phi c)' W (y - \Phi c) + \lambda c'Rc

    where :math:`y` is the vector or matrix of observations, :math:`\Phi`
    the matrix whose columns are the basis functions evaluated at the
    sampling points, :math:`c` the coefficient vector or matrix to be
    estimated, :math:`\lambda` a smoothness parameter and :math:`c'Rc` the
    matrix representation of the roughness penalty :math:`\int \left[ L(
    x(s)) \right] ^2 ds` where :math:`L` is a linear differential operator.

    Each element of :math:`R` has the following close form:

    .. math::

        R_{ij} = \int L\phi_i(s) L\phi_j(s) ds

    By deriving the first formula we obtain the closed formed of the
    estimated coefficients matrix:

    .. math::

        \hat{c} = \left( \Phi' W \Phi + \lambda R \right)^{-1} \Phi' W y

    The solution of this matrix equation is done using the cholesky
    method for the resolution of a LS problem. If this method throughs a
    rounding error warning you may want to use the QR factorisation that
    is more numerically stable despite being more expensive to compute.
    [RS05-5-2-8]_

    Args:
        basis: Basis used.
        weights: Matrix to weight the observations. Defaults to the identity
            matrix.
        smoothing_parameter: Smoothing parameter. Trying with several
            factors in a logarithm scale is suggested. If 0 no smoothing is
            performed. Defaults to 1.
        regularization: Regularization object. This allows the penalization of
            complicated models, which applies additional smoothing. By default
            is ``None`` meaning that no additional smoothing has to take
            place.
        method: Algorithm used for calculating the coefficients using
            the least squares method. The values admitted are 'cholesky', 'qr'
            and 'svd' for Cholesky, QR and SVD factorisation methods
            respectively, or a callable similar to the `lstsq` function. The
            default is 'svd', which is the most robust but less performant one.
        output_points: The output points. If ommited, the input points are
            used. If ``return_basis`` is ``True``, this parameter is ignored.
        return_basis: If ``False`` (the default) returns the smoothed
            data as an FDataGrid, like the other smoothers. If ``True`` returns
            a FDataBasis object.

    Examples:
        By default, this smoother returns a FDataGrid, like the other
        smoothers:

        >>> import numpy as np
        >>> import skfda
        >>> t = np.linspace(0, 1, 5)
        >>> x = np.sin(2 * np.pi * t) + np.cos(2 * np.pi * t) + 2
        >>> x
        array([ 3.,  3.,  1.,  1.,  3.])

        >>> fd = skfda.FDataGrid(data_matrix=x, grid_points=t)
        >>> basis = skfda.representation.basis.Fourier((0, 1), n_basis=3)
        >>> smoother = skfda.preprocessing.smoothing.BasisSmoother(basis)
        >>> fd_smooth = smoother.fit_transform(fd)
        >>> fd_smooth.data_matrix.round(2)
        array([[[ 3.],
                [ 3.],
                [ 1.],
                [ 1.],
                [ 3.]]])

        However, the parameter ``return_basis`` can be used to return the data
        in basis form, by default, without extra smoothing:

        >>> fd = skfda.FDataGrid(data_matrix=x, grid_points=t)
        >>> basis = skfda.representation.basis.Fourier((0, 1), n_basis=3)
        >>> smoother = skfda.preprocessing.smoothing.BasisSmoother(
        ...     basis,
        ...     method='cholesky',
        ...     return_basis=True,
        ... )
        >>> fd_basis = smoother.fit_transform(fd)
        >>> fd_basis.coefficients.round(2)
        array([[ 2.  , 0.71, 0.71]])

        >>> smoother = skfda.preprocessing.smoothing.BasisSmoother(
        ...     basis,
        ...     method='qr',
        ...     return_basis=True,
        ... )
        >>> fd_basis = smoother.fit_transform(fd)
        >>> fd_basis.coefficients.round(2)
        array([[ 2.  , 0.71, 0.71]])

        >>> smoother = skfda.preprocessing.smoothing.BasisSmoother(
        ...     basis,
        ...     method='svd',
        ...     return_basis=True,
        ... )
        >>> fd_basis = smoother.fit_transform(fd)
        >>> fd_basis.coefficients.round(2)
        array([[ 2.  , 0.71, 0.71]])
        >>> smoother.hat_matrix().round(2)
        array([[ 0.43,  0.14, -0.14,  0.14,  0.43],
               [ 0.14,  0.71,  0.29, -0.29,  0.14],
               [-0.14,  0.29,  0.71,  0.29, -0.14],
               [ 0.14, -0.29,  0.29,  0.71,  0.14],
               [ 0.43,  0.14, -0.14,  0.14,  0.43]])

        We can penalize approximations that are not smooth enough using some
        kind of regularization:

        >>> from skfda.misc.regularization import L2Regularization
        >>> from skfda.misc.operators import LinearDifferentialOperator
        >>>
        >>> fd = skfda.FDataGrid(data_matrix=x, grid_points=t)
        >>> basis = skfda.representation.basis.Fourier((0, 1), n_basis=3)
        >>> smoother = skfda.preprocessing.smoothing.BasisSmoother(
        ...     basis,
        ...     method='cholesky',
        ...     regularization=L2Regularization(
        ...         LinearDifferentialOperator([0.1, 0.2]),
        ...     ),
        ...     return_basis=True,
        ... )
        >>> fd_basis = smoother.fit_transform(fd)
        >>> fd_basis.coefficients.round(2)
        array([[ 2.04,  0.51,  0.55]])

        >>> fd = skfda.FDataGrid(data_matrix=x, grid_points=t)
        >>> basis = skfda.representation.basis.Fourier((0, 1), n_basis=3)
        >>> smoother = skfda.preprocessing.smoothing.BasisSmoother(
        ...     basis,
        ...     method='qr',
        ...     regularization=L2Regularization(
        ...         LinearDifferentialOperator([0.1, 0.2]),
        ...     ),
        ...     return_basis=True,
        ... )
        >>> fd_basis = smoother.fit_transform(fd)
        >>> fd_basis.coefficients.round(2)
        array([[ 2.04,  0.51,  0.55]])

        >>> fd = skfda.FDataGrid(data_matrix=x, grid_points=t)
        >>> basis = skfda.representation.basis.Fourier((0, 1), n_basis=3)
        >>> smoother = skfda.preprocessing.smoothing.BasisSmoother(
        ...     basis,
        ...     method='svd',
        ...     regularization=L2Regularization(
        ...         LinearDifferentialOperator([0.1, 0.2]),
        ...     ),
        ...     return_basis=True,
        ... )
        >>> fd_basis = smoother.fit_transform(fd)
        >>> fd_basis.coefficients.round(2)
        array([[ 2.04,  0.51,  0.55]])

    References:
        .. [RS05-5-2-6] Ramsay, J., Silverman, B. W. (2005). How spline
            smooths are computed. In *Functional Data Analysis*
            (pp. 86-87). Springer.

        .. [RS05-5-2-8] Ramsay, J., Silverman, B. W. (2005). HSpline
            smoothing as an augmented least squares problem. In *Functional
            Data Analysis* (pp. 86-87). Springer.

    """

    _required_parameters = ["basis"]

    def __init__(
        self,
        basis: Basis,
        *,
        smoothing_parameter: float = 1.0,
        weights: Optional[NDArrayFloat] = None,
        regularization: Optional[L2Regularization[FDataGrid]] = None,
        output_points: Optional[GridPointsLike] = None,
        method: LstsqMethod = 'svd',
        return_basis: bool = False,
    ) -> None:
        self.basis = basis
        self.smoothing_parameter = smoothing_parameter
        self.weights = weights
        self.regularization = regularization
        self.output_points = output_points
        self.method = method
        self.return_basis: Final = return_basis

    def _coef_matrix(
        self,
        input_points: GridPointsLike,
        *,
        data_matrix: Optional[NDArrayFloat] = None,
    ) -> NDArrayFloat:
        """Get the matrix that gives the coefficients."""
        from ...misc.regularization import compute_penalty_matrix

        basis_values_input = self.basis.evaluate(
            _cartesian_product(_to_grid_points(input_points)),
        ).reshape((self.basis.n_basis, -1)).T

        penalty_matrix = compute_penalty_matrix(
            basis_iterable=(self.basis,),
            regularization_parameter=self.smoothing_parameter,
            regularization=self.regularization,
        )

        # Get the matrix for computing the coefficients if no
        # data_matrix is passed
        if data_matrix is None:
            data_matrix = np.eye(basis_values_input.shape[0])

        return solve_regularized_weighted_lstsq(
            coefs=basis_values_input,
            result=data_matrix,
            weights=self.weights,
            penalty_matrix=penalty_matrix,
            lstsq_method=self.method,
        )

    def _hat_matrix(
        self,
        input_points: GridPointsLike,
        output_points: GridPointsLike,
    ) -> NDArrayFloat:
        basis_values_output = self.basis.evaluate(
            _cartesian_product(
                _to_grid_points(output_points),
            ),
        ).reshape((self.basis.n_basis, -1)).T

        return basis_values_output @ self._coef_matrix(input_points)

    def fit(
        self,
        X: FDataGrid,
        y: None = None,
    ) -> BasisSmoother:
        """Compute the hat matrix for the desired output points.

        Args:
            X: The data whose points are used to compute the matrix.
            y: Ignored.

        Returns:
            self

        """
        self.input_points_ = X.grid_points
        self.output_points_ = (
            _to_grid_points(self.output_points)
            if self.output_points is not None
            else self.input_points_
        )

        if not self.return_basis:
            super().fit(X, y)

        return self

    def transform(
        self,
        X: FDataGrid,
        y: None = None,
    ) -> FData:
        """
        Smooth the data.

        Args:
            X: The data to smooth.
            y: Ignored

        Returns:
            Smoothed data.

        """
        assert all(
            np.array_equal(i, s) for i, s in zip(
                self.input_points_,
                X.grid_points,
            )
        )

        if self.return_basis:
            coefficients = self._coef_matrix(
                input_points=X.grid_points,
                data_matrix=X.data_matrix.reshape((X.n_samples, -1)).T,
            ).T

            return FDataBasis(
                basis=self.basis,
                coefficients=coefficients,
                dataset_name=X.dataset_name,
                argument_names=X.argument_names,
                coordinate_names=X.coordinate_names,
                sample_names=X.sample_names,
            )

        return super().transform(X, y)



from __future__ import annotations

import math
from typing import Tuple, Union

import numpy as np
import scipy.integrate
from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.utils.validation import check_is_fitted

from ..._utils import _cartesian_product, _pairwise_symmetric
from ...representation import FDataBasis, FDataGrid
from ...representation.basis import Basis, FiniteElement, VectorValued

_MeanType = Union[FDataGrid, float]


def _pairwise_fem_inner_product(
    basis_fd: FDataBasis,
    fd: FDataGrid,
    y_val: float,
    grid: np.ndarray,
) -> np.ndarray:

    eval_grid_fem = np.concatenate(
        (
            grid[:, None],
            np.full(
                shape=(len(grid), 1),
                fill_value=y_val,
            ),
        ),
        axis=1,
    )

    eval_fem = basis_fd(eval_grid_fem)
    eval_fd = fd(grid)

    prod = eval_fem * eval_fd
    integral = scipy.integrate.simps(prod, grid, axis=1)
    return np.sum(integral, axis=-1)


def _inner_product_matrix(
    basis: Basis,
    fd: FDataGrid,
    limits: Tuple[float, float],
    y_val: float,
) -> np.ndarray:
    """
    Compute inner products with the FEM basis.

    Compute the matrix of inner products of an FEM basis with a functional
    data object over a range of x-values for a fixed y-value. The numerical
    integration uses Romberg integration with the trapezoidal rule.

    Arguments:
        basis: typically a FEM basis defined by a triangulation within a
            rectangular domain. It is assumed that only the part of the mesh
            that is within the upper left triangular is of interest.
        fd: a regular functional data object.
        limits: limits of integration, as a tuple of form
            (lower limit, upper limit)
        y_val: the fixed y value.

    Returns:
        Matrix of inner products.

    """
    basis_fd = basis.to_basis()
    grid = fd.grid_points[0]
    grid_index = (grid >= limits[0]) & (grid <= limits[1])
    grid = grid[grid_index]

    return _pairwise_symmetric(
        _pairwise_fem_inner_product,
        basis_fd,
        fd,
        y_val=y_val,
        grid=grid,
    )


def _design_matrix(
    basis: Basis,
    fd: FDataGrid,
    pred_points: np.ndarray,
) -> np.ndarray:
    """
    Compute the indefinite integrals of the curves over s up to each t-value.

    Arguments:
        basis: typically a FEM basis defined by a triangulation within a
            rectangular domain. It is assumed that only the part of the mesh
            that is within the upper left triangular is of interest.
        fd: a regular functional data object.
        pred_points: points where ``fd`` is evaluated.

    Returns:
        Design matrix.

    """
    matrix = np.array([
        _inner_product_matrix(basis, fd, limits=(0, t), y_val=t).T
        for t in pred_points
    ])

    return np.swapaxes(matrix, 0, 1)


def _get_valid_points(
    interval_len: float,
    n_intervals: int,
    lag: float,
) -> np.ndarray:
    """Return the valid points as integer tuples."""
    interval_points = np.arange(n_intervals + 1)
    full_grid_points = _cartesian_product((interval_points, interval_points))

    past_points = full_grid_points[
        full_grid_points[:, 0] <= full_grid_points[:, 1]
    ]

    discrete_lag = np.inf if lag == np.inf else math.ceil(lag / interval_len)

    return past_points[
        past_points[:, 1] - past_points[:, 0] <= discrete_lag
    ]


def _get_triangles(
    n_intervals: int,
    valid_points: np.ndarray,
) -> np.ndarray:
    """Construct the triangle grid given the valid points."""
    # A matrix where the (integer) coords of a point match
    # to its index or to -1 if it does not exist.
    indexes_matrix = np.full(
        shape=(n_intervals + 1, n_intervals + 1),
        fill_value=-1,
        dtype=np.int_,
    )

    indexes_matrix[
        valid_points[:, 0],
        valid_points[:, 1],
    ] = np.arange(len(valid_points))

    interval_without_end = np.arange(n_intervals)

    pts_coords = _cartesian_product(
        (interval_without_end, interval_without_end),
    )

    pts_coords_x = pts_coords[:, 0]
    pts_coords_y = pts_coords[:, 1]

    down_triangles = np.stack(
        (
            indexes_matrix[pts_coords_x, pts_coords_y],
            indexes_matrix[pts_coords_x + 1, pts_coords_y],
            indexes_matrix[pts_coords_x + 1, pts_coords_y + 1],
        ),
        axis=1,
    )

    up_triangles = np.stack(
        (
            indexes_matrix[pts_coords_x, pts_coords_y],
            indexes_matrix[pts_coords_x, pts_coords_y + 1],
            indexes_matrix[pts_coords_x + 1, pts_coords_y + 1],
        ),
        axis=1,
    )

    triangles = np.concatenate((down_triangles, up_triangles))
    has_wrong_index = np.any(triangles < 0, axis=1)

    return triangles[~has_wrong_index]


def _create_fem_basis(
    start: float,
    stop: float,
    n_intervals: int,
    lag: float,
) -> FiniteElement:

    interval_len = (stop - start) / n_intervals

    valid_points = _get_valid_points(
        interval_len=interval_len,
        n_intervals=n_intervals,
        lag=lag,
    )

    final_points = valid_points * interval_len + start

    triangles = _get_triangles(
        n_intervals=n_intervals,
        valid_points=valid_points,
    )

    return FiniteElement(
        vertices=final_points,
        cells=triangles,
        domain_range=((start, stop),) * 2,
    )


class HistoricalLinearRegression(
    BaseEstimator,  # type: ignore
    RegressorMixin,  # type: ignore
):
    r"""Historical functional linear regression.

    This is a linear regression method where the covariate and the response are
    both functions :math:`\mathbb{R}` to :math:`\mathbb{R}` with the same
    domain. In order to predict the value of the response function at point
    :math:`t`, only the information of the covariate at points :math:`s < t` is
    used. Is thus an "historical" model in the sense that, if the domain
    represents time, only the data from the past, or historical data, is used
    to predict a given point.

    The model assumed by this method is:

    .. math::
        y_i = \alpha(t) + \int_{s_0(t)}^t x_i(s) \beta(s, t) ds

    where :math:`s_0(t) = \max(0, t - \delta)` and :math:`\delta` is a
    predefined time lag that can be specified so that points far in the past
    do not affect the predicted value.

    Args:
        n_intervals: Number of intervals used to create the basis of the
            coefficients. This will be a bidimensional
            :class:`~skfda.representation.basis.FiniteElement` basis, and
            this parameter indirectly specifies the number of
            elements of that basis, and thus the granularity.
        fit_intercept:  Whether to calculate the intercept for this
            model. If set to False, no intercept will be used in calculations
            (i.e. data is expected to be centered).
        lag: The maximum time lag at which points in the past can still
            influence the prediction.

    Attributes:
        basis_coef\_: The fitted coefficient function as a FDataBasis.
        coef\_: The fitted coefficient function as a FDataGrid.
        intercept\_: Independent term in the linear model. Set to the constant
            function 0 if `fit_intercept = False`.

    Examples:
        The following example test a case that conforms to this model.

        >>> from skfda import FDataGrid
        >>> from skfda.ml.regression import HistoricalLinearRegression
        >>> import numpy as np
        >>> import scipy.integrate

        >>> random_state = np.random.RandomState(0)
        >>> data_matrix = random_state.choice(10, size=(8, 6))
        >>> data_matrix
        array([[5, 0, 3, 3, 7, 9],
               [3, 5, 2, 4, 7, 6],
               [8, 8, 1, 6, 7, 7],
               [8, 1, 5, 9, 8, 9],
               [4, 3, 0, 3, 5, 0],
               [2, 3, 8, 1, 3, 3],
               [3, 7, 0, 1, 9, 9],
               [0, 4, 7, 3, 2, 7]])
        >>> intercept = random_state.choice(10, size=(1, 6))
        >>> intercept
        array([[2, 0, 0, 4, 5, 5]])
        >>> y_data = scipy.integrate.cumtrapz(
        ...              data_matrix,
        ...              initial=0,
        ...              axis=1,
        ...          ) + intercept
        >>> y_data
        array([[  2. ,   2.5,   4. ,  11. ,  17. ,  25. ],
               [  2. ,   4. ,   7.5,  14.5,  21. ,  27.5],
               [  2. ,   8. ,  12.5,  20. ,  27.5,  34.5],
               [  2. ,   4.5,   7.5,  18.5,  28. ,  36.5],
               [  2. ,   3.5,   5. ,  10.5,  15.5,  18. ],
               [  2. ,   2.5,   8. ,  16.5,  19.5,  22.5],
               [  2. ,   5. ,   8.5,  13. ,  19. ,  28. ],
               [  2. ,   2. ,   7.5,  16.5,  20. ,  24.5]])
        >>> X = FDataGrid(data_matrix)
        >>> y = FDataGrid(y_data)
        >>> hist = HistoricalLinearRegression(n_intervals=8)
        >>> _ = hist.fit(X, y)
        >>> hist.predict(X).data_matrix[..., 0].round(1)
        array([[  2. ,   2.5,   4. ,  11. ,  17. ,  25. ],
               [  2. ,   4. ,   7.5,  14.5,  21. ,  27.5],
               [  2. ,   8. ,  12.5,  20. ,  27.5,  34.5],
               [  2. ,   4.5,   7.5,  18.5,  28. ,  36.5],
               [  2. ,   3.5,   5. ,  10.5,  15.5,  18. ],
               [  2. ,   2.5,   8. ,  16.5,  19.5,  22.5],
               [  2. ,   5. ,   8.5,  13. ,  19. ,  28. ],
               [  2. ,   2. ,   7.5,  16.5,  20. ,  24.5]])
        >>> abs(hist.intercept_.data_matrix[..., 0].round())
        array([[ 2.,  0.,  0.,  4.,  5.,  5.]])

    References:
        Malfait, N., & Ramsay, J. O. (2003). The historical functional linear
        model. Canadian Journal of Statistics, 31(2), 115-128.

    """

    def __init__(
        self,
        *,
        n_intervals: int,
        fit_intercept: bool = True,
        lag: float = math.inf,
    ) -> None:
        self.n_intervals = n_intervals
        self.fit_intercept = fit_intercept
        self.lag = lag

    def _center_X_y(
        self,
        X: FDataGrid,
        y: FDataGrid,
    ) -> Tuple[FDataGrid, FDataGrid, _MeanType, _MeanType]:

        X_mean: Union[FDataGrid, float] = (
            X.mean() if self.fit_intercept else 0
        )
        X_centered = X - X_mean
        y_mean: Union[FDataGrid, float] = (
            y.mean() if self.fit_intercept else 0
        )
        y_centered = y - y_mean

        return X_centered, y_centered, X_mean, y_mean

    def _fit_and_return_centered_matrix(
        self,
        X: FDataGrid,
        y: FDataGrid,
    ) -> Tuple[np.ndarray, _MeanType]:

        X_centered, y_centered, X_mean, y_mean = self._center_X_y(X, y)

        self._pred_points = y_centered.grid_points[0]
        self._pred_domain_range = y_centered.domain_range[0]

        fem_basis = _create_fem_basis(
            start=X_centered.domain_range[0][0],
            stop=X_centered.domain_range[0][1],
            n_intervals=self.n_intervals,
            lag=self.lag,
        )

        self._basis = VectorValued(
            [fem_basis] * X_centered.dim_codomain
        )

        design_matrix = _design_matrix(
            self._basis,
            X_centered,
            pred_points=self._pred_points,
        )
        design_matrix = design_matrix.reshape(-1, design_matrix.shape[-1])

        self._coef_coefs = np.linalg.lstsq(
            design_matrix,
            y_centered.data_matrix[:, ..., 0].ravel(),
            rcond=None,
        )[0]

        self.basis_coef_ = FDataBasis(
            basis=self._basis,
            coefficients=self._coef_coefs,
        )

        self.coef_ = self.basis_coef_.to_grid(
            grid_points=[X.grid_points[0]] * 2,
        )

        if self.fit_intercept:
            assert isinstance(X_mean, FDataGrid)
            self.intercept_ = (
                y_mean - self._predict_no_intercept(X_mean)
            )
        else:
            self.intercept_ = y.copy(
                data_matrix=np.zeros_like(y.data_matrix[0]),
            )

        return design_matrix, y_mean

    def _prediction_from_matrix(self, design_matrix: np.ndarray) -> FDataGrid:

        points = (design_matrix @ self._coef_coefs).reshape(
            -1,
            len(self._pred_points),
        )

        return FDataGrid(
            points,
            grid_points=self._pred_points,
            domain_range=self._pred_domain_range,
        )

    def fit(  # noqa: D102
        self,
        X: FDataGrid,
        y: FDataGrid,
    ) -> HistoricalLinearRegression:

        self._fit_and_return_centered_matrix(X, y)
        return self

    def fit_predict(  # noqa: D102
        self,
        X: FDataGrid,
        y: FDataGrid,
    ) -> FDataGrid:

        design_matrix, y_mean = self._fit_and_return_centered_matrix(X, y)
        return (
            self._prediction_from_matrix(design_matrix)
            + y_mean
        )

    def _predict_no_intercept(self, X: FDataGrid) -> FDataGrid:

        design_matrix = _design_matrix(
            self._basis,
            X,
            pred_points=self._pred_points,
        )
        design_matrix = design_matrix.reshape(-1, design_matrix.shape[-1])

        return self._prediction_from_matrix(design_matrix)

    def predict(self, X: FDataGrid) -> FDataGrid:  # noqa: D102

        check_is_fitted(self)

        return self._predict_no_intercept(X) + self.intercept_



from typing import Optional, TypeVar

import numpy as np

from .._typing import ArrayLike, DomainRangeLike
from ._basis import Basis

T = TypeVar("T", bound='FiniteElement')


class FiniteElement(Basis):
    """Finite element basis.

    Given a n-dimensional grid made of simplices, each element of the basis
    is a piecewise linear function that takes the value 1 at exactly one
    vertex and 0 in the other vertices.

    Parameters:
        vertices: The vertices of the grid.
        cells: A list of individual cells, consisting in the indexes of
            :math:`n+1` vertices for an n-dimensional domain space.

    Examples:
        >>> from skfda.representation.basis import FiniteElement
        >>> basis = FiniteElement(
        ...     vertices=[[0, 0], [0, 1], [1, 0], [1, 1]],
        ...     cells=[[0, 1, 2], [1, 2, 3]],
        ...     domain_range=[(0, 1), (0, 1)],
        ... )

        Evaluates all the functions in the basis in a list of discrete
        values.

        >>> basis([[0.1, 0.1], [0.6, 0.6], [0.1, 0.2], [0.8, 0.9]])
        array([[[ 0.8],
                [ 0. ],
                [ 0.7],
                [ 0. ]],
               [[ 0.1],
                [ 0.4],
                [ 0.2],
                [ 0.2]],
               [[ 0.1],
                [ 0.4],
                [ 0.1],
                [ 0.1]],
               [[ 0. ],
                [ 0.2],
                [ 0. ],
                [ 0.7]]])


        >>> from scipy.spatial import Delaunay
        >>> import numpy as np
        >>>
        >>> n_points = 10
        >>> points = np.random.uniform(size=(n_points, 2))
        >>> delaunay = Delaunay(points)
        >>> basis = FiniteElement(
        ...     vertices=delaunay.points,
        ...     cells=delaunay.simplices,
        ... )
        >>> basis.n_basis
        10

    """

    def __init__(
        self,
        vertices: ArrayLike,
        cells: ArrayLike,
        domain_range: Optional[DomainRangeLike] = None,
    ) -> None:
        super().__init__(
            domain_range=domain_range,
            n_basis=len(vertices),
        )
        self.vertices = np.asarray(vertices)
        self.cells = np.asarray(cells)

    @property
    def dim_domain(self) -> int:
        return self.vertices.shape[-1]

    def _barycentric_coords(self, points: np.ndarray) -> np.ndarray:
        """
        Find the barycentric coordinates of each point in each cell.

        Only works for simplex cells.

        """
        cell_coordinates = self.vertices[self.cells]

        cartesian_matrix = np.append(
            cell_coordinates,
            np.ones(cell_coordinates.shape[:-1] + (1,)),
            axis=-1,
        )

        cartesian_vector = np.append(
            points,
            np.ones(points.shape[:-1] + (1,)),
            axis=-1,
        )

        coords = np.linalg.solve(
            np.swapaxes(cartesian_matrix, -2, -1),
            cartesian_vector.T[np.newaxis, ...],
        )

        return np.swapaxes(coords, -2, -1)

    def _cell_points_values(self, points: np.ndarray) -> np.ndarray:
        """
        Compute the values of each point in each of the vertices of each cell.

        Only works for simplex cells.

        """
        barycentric_coords = self._barycentric_coords(points)

        # Remove values outside each cell
        wrong_vals = np.any(
            ((barycentric_coords < 0) & ~np.isclose(barycentric_coords + 1, 1))
            | ((barycentric_coords > 1) & ~np.isclose(barycentric_coords, 1)),
            axis=-1,
        )

        barycentric_coords[wrong_vals] = 0

        points_in_cells = np.any(barycentric_coords, axis=-1)
        n_cells_per_point = np.sum(points_in_cells, axis=0)

        n_cells_per_point[n_cells_per_point == 0] = 1
        barycentric_coords /= n_cells_per_point[:, np.newaxis]

        return barycentric_coords

    def _evaluate(self, eval_points: np.ndarray) -> np.ndarray:

        points_values_per_cell = self._cell_points_values(eval_points)

        cell_points_values = np.swapaxes(points_values_per_cell, -2, -1)
        cell_points_values = cell_points_values.reshape(-1, len(eval_points))
        indexes = self.cells.ravel()

        eval_matrix = np.zeros((self.n_basis, len(eval_points)))
        np.add.at(eval_matrix, indexes, cell_points_values)

        return eval_matrix



import copy
import itertools
from functools import partial
from typing import (
    Generator,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
    cast,
)

import numpy as np
from matplotlib.artist import Artist
from matplotlib.axes import Axes
from matplotlib.backend_bases import Event, LocationEvent, MouseEvent
from matplotlib.collections import PathCollection
from matplotlib.figure import Figure
from matplotlib.text import Annotation
from matplotlib.widgets import Slider, Widget

from ._baseplot import BasePlot
from ._utils import _get_axes_shape, _get_figure_and_axes, _set_figure_layout


def _set_val_noevents(widget: Widget, val: float) -> None:
    e = widget.eventson
    widget.eventson = False
    widget.set_val(val)
    widget.eventson = e


class MultipleDisplay:
    """
    MultipleDisplay class used to combine and interact with plots.

    This module is used to combine different BasePlot objects that
    represent the same curves or surfaces, and represent them
    together in the same figure. Besides this, it includes
    the functionality necessary to interact with the graphics
    by clicking the points, hovering over them... Picking the points allow
    us to see our selected function standing out among the others in all
    the axes. It is also possible to add widgets to interact with the
    plots.
    Args:
        displays: Baseplot objects that will be plotted in the fig.
        criteria: Sequence of criteria used to order the points in the
            slider widget. The size should be equal to sliders, as each
            criterion is for one slider.
        sliders: Sequence of widgets that will be plotted.
        label_sliders: Label of each of the sliders.
        chart: Figure over with the graphs are plotted or axis over
            where the graphs are plotted. If None and ax is also
            None, the figure is initialized.
        fig: Figure over with the graphs are plotted in case ax is not
            specified. If None and ax is also None, the figure is
            initialized.
        axes: Axis where the graphs are plotted. If None, see param fig.
    Attributes:
        length_data: Number of instances or curves of the different displays.
        clicked: Boolean indicating whether a point has being clicked.
        selected_sample: Index of the function selected with the interactive
            module or widgets.
    """

    def __init__(
        self,
        displays: Union[BasePlot, Sequence[BasePlot]],
        criteria: Union[Sequence[float], Sequence[Sequence[float]]] = (),
        sliders: Union[Type[Widget], Sequence[Type[Widget]]] = (),
        label_sliders: Union[str, Sequence[str], None] = None,
        chart: Union[Figure, Axes, None] = None,
        fig: Optional[Figure] = None,
        axes: Optional[Sequence[Axes]] = None,
    ):
        if isinstance(displays, BasePlot):
            displays = (displays,)

        self.displays = [copy.copy(d) for d in displays]
        self._n_graphs = sum(d.n_subplots for d in self.displays)
        self.length_data = next(
            d.n_samples
            for d in self.displays
            if d.n_samples is not None
        )
        self.sliders: List[Widget] = []
        self.selected_sample: Optional[int] = None

        if len(criteria) != 0 and not isinstance(criteria[0], Sequence):
            criteria = (criteria,)

        criteria = cast(Sequence[Sequence[float]], criteria)
        self.criteria = criteria

        if not isinstance(sliders, Sequence):
            sliders = (sliders,)

        if isinstance(label_sliders, str):
            label_sliders = (label_sliders,)

        if len(criteria) != len(sliders):
            raise ValueError(
                f"Size of criteria, and sliders should be equal "
                f"(have {len(criteria)} and {len(sliders)}).",
            )

        self._init_axes(
            chart,
            fig=fig,
            axes=axes,
            extra=len(criteria),
        )

        self._create_sliders(
            criteria=criteria,
            sliders=sliders,
            label_sliders=label_sliders,
        )

    def _init_axes(
        self,
        chart: Union[Figure, Axes, None] = None,
        *,
        fig: Optional[Figure] = None,
        axes: Optional[Sequence[Axes]] = None,
        extra: int = 0,
    ) -> None:
        """
        Initialize the axes and figure.

        Args:
            chart: Figure over with the graphs are plotted or axis over
                where the graphs are plotted. If None and ax is also
                None, the figure is initialized.
            fig: Figure over with the graphs are plotted in case ax is not
                specified. If None and ax is also None, the figure is
                initialized.
            axes: Axis where the graphs are plotted. If None, see param fig.
            extra: integer indicating the extra axes needed due to the
                necessity for them to plot the sliders.

        """
        widget_aspect = 1 / 8
        fig, axes = _get_figure_and_axes(chart, fig, axes)
        if len(axes) not in {0, self._n_graphs + extra}:
            raise ValueError("Invalid number of axes.")

        n_rows, n_cols = _get_axes_shape(self._n_graphs + extra)

        dim = list(
            itertools.chain.from_iterable(
                [d.dim] * d.n_subplots
                for d in self.displays
            ),
        ) + [2] * extra

        number_axes = n_rows * n_cols
        fig, axes = _set_figure_layout(
            fig=fig,
            axes=axes,
            n_axes=self._n_graphs + extra,
            dim=dim,
        )

        for i in range(self._n_graphs, number_axes):
            if i >= self._n_graphs + extra:
                axes[i].set_visible(False)
            else:
                axes[i].set_box_aspect(widget_aspect)

        self.fig = fig
        self.axes = axes

    def _create_sliders(
        self,
        *,
        criteria: Sequence[Sequence[float]],
        sliders: Sequence[Type[Widget]],
        label_sliders: Optional[Sequence[str]] = None,
    ) -> None:
        """
        Create the sliders with the criteria selected.

        Args:
            criteria: Different criterion for each of the sliders.
            sliders: Widget types.
            label_sliders: Sequence of the names of each slider.

        """
        for c in criteria:
            if len(c) != self.length_data:
                raise ValueError(
                    "Slider criteria should be of the same size as data",
                )

        for k, criterion in enumerate(criteria):
            label = label_sliders[k] if label_sliders else None

            self.add_slider(
                axes=self.axes[self._n_graphs + k],
                criterion=criterion,
                widget_class=sliders[k],
                label=label,
            )

    def plot(self) -> Figure:
        """
        Plot Multiple Display method.

        Plot the different BasePlot objects and widgets selected.
        Activates the interactivity functionality of clicking and
        hovering points. When clicking a point, the rest will be
        made partially transparent in all the corresponding graphs.
        Returns:
            fig: figure object in which the displays and
                widgets will be plotted.
        """
        if self._n_graphs > 1:
            for d in self.displays[1:]:
                if (
                    d.n_samples is not None
                    and d.n_samples != self.length_data
                ):
                    raise ValueError(
                        "Length of some data sets are not equal ",
                    )

        for ax in self.axes[:self._n_graphs]:
            ax.clear()

        int_index = 0
        for disp in self.displays:
            axes_needed = disp.n_subplots
            end_index = axes_needed + int_index
            disp._set_figure_and_axes(axes=self.axes[int_index:end_index])
            disp.plot()
            int_index = end_index

        self.fig.canvas.mpl_connect('pick_event', self.pick)

        self.fig.suptitle("Multiple display")
        self.fig.tight_layout()

        return self.fig

    def pick(self, event: Event) -> None:
        """
        Activate interactive functionality when picking a point.

        Callback method that is activated when a point is picked.
        If no point was clicked previously, all the points but the
        one selected will be more transparent in all the graphs.
        If a point was clicked already, this new point will be the
        one highlighted among the rest. If the same point is clicked,
        the initial state of the graphics is restored.
        Args:
            event: event object containing the artist of the point
                picked.
        """
        selected_sample = self._sample_from_artist(event.artist)

        if selected_sample is not None:
            if self.selected_sample == selected_sample:
                self._deselect_samples()
            else:
                self._select_sample(selected_sample)

    def _sample_from_artist(self, artist: Artist) -> Optional[int]:
        """Return the sample corresponding to an artist."""
        for d in self.displays:

            if d.artists is None:
                continue

            for i, a in enumerate(d.axes_):
                if a == artist.axes:
                    if len(d.axes_) == 1:
                        return np.where(d.artists == artist)[0][0]
                    else:
                        return np.where(d.artists[:, i] == artist)[0][0]

        return None

    def _visit_artists(self) -> Generator[Tuple[int, Artist], None, None]:
        for i in range(self.length_data):
            for d in self.displays:
                if d.artists is None:
                    continue

                yield from ((i, artist) for artist in np.ravel(d.artists[i]))

    def _select_sample(self, selected_sample: int) -> None:
        """Reduce the transparency of all the points but the selected one."""
        for i, artist in self._visit_artists():
            artist.set_alpha(1.0 if i == selected_sample else 0.1)

        for criterion, slider in zip(self.criteria, self.sliders):
            val_widget = criterion[selected_sample]
            _set_val_noevents(slider, val_widget)

        self.selected_sample = selected_sample
        self.fig.canvas.draw_idle()

    def _deselect_samples(self) -> None:
        """Restore the original transparency of all the points."""
        for _, artist in self._visit_artists():
            artist.set_alpha(1)

        self.selected_sample = None
        self.fig.canvas.draw_idle()

    def add_slider(
        self,
        axes: Axes,
        criterion: Sequence[float],
        widget_class: Type[Widget] = Slider,
        label: Optional[str] = None,
    ) -> None:
        """
        Add the slider to the MultipleDisplay object.

        Args:
            axes: Axes for the widget.
            criterion: Criterion used for the slider.
            widget_class: Widget type.
            label: Name of the slider.
        """
        full_desc = "" if label is None else label

        ordered_criterion_values, ordered_criterion_indexes = zip(
            *sorted(zip(criterion, range(self.length_data))),
        )

        widget = widget_class(
            ax=axes,
            label=full_desc,
            valmin=ordered_criterion_values[0],
            valmax=ordered_criterion_values[-1],
            valinit=ordered_criterion_values[0],
            valstep=ordered_criterion_values,
            valfmt="%.3g",
        )

        self.sliders.append(widget)

        axes.annotate(
            f"{ordered_criterion_values[0]:.3g}",
            xy=(0, -0.5),
            xycoords='axes fraction',
            annotation_clip=False,
        )

        axes.annotate(
            f"{ordered_criterion_values[-1]:.3g}",
            xy=(0.95, -0.5),
            xycoords='axes fraction',
            annotation_clip=False,
        )

        on_changed_function = partial(
            self._value_updated,
            ordered_criterion_values=ordered_criterion_values,
            ordered_criterion_indexes=ordered_criterion_indexes,
        )

        widget.on_changed(on_changed_function)

    def _value_updated(
        self,
        value: float,
        ordered_criterion_values: Sequence[float],
        ordered_criterion_indexes: Sequence[int],
    ) -> None:
        """
        Update the graphs when a widget is clicked.

        Args:
            value: Current value of the widget.
            ordered_criterion_values: Ordered values of the criterion.
            ordered_criterion_indexes: Sample numbers ordered using the
                criterion.

        """
        value_index = int(np.searchsorted(ordered_criterion_values, value))
        self.selected_sample = ordered_criterion_indexes[value_index]
        self._select_sample(self.selected_sample)



from __future__ import annotations

from typing import Any, Optional

import numpy as np
import scipy.integrate
from fdasrsf.utility_functions import optimum_reparam

from ..._utils import check_is_univariate, invert_warping, normalize_scale
from ...misc.operators import SRSF
from ...representation import FDataGrid
from ...representation._typing import NDArrayFloat
from ...representation.interpolation import SplineInterpolation

###############################################################################
# Based on the original implementation of J. Derek Tucker in                  #
# *fdasrsf_python* (https://github.com/jdtuck/fdasrsf_python)                 #
# and *ElasticFDA.jl* (https://github.com/jdtuck/ElasticFDA.jl).              #
###############################################################################


def _elastic_alignment_array(
    template_data: NDArrayFloat,
    q_data: NDArrayFloat,
    eval_points: NDArrayFloat,
    penalty: float,
    grid_dim: int,
) -> NDArrayFloat:
    """
    Wrap the :func:`optimum_reparam` function of fdasrsf.

    Selects the corresponding routine depending on the dimensions of the
    arrays.

    Args:
        template_data: Array with the srsf of the template.
        q_data: Array with the srsf of the curves
                to be aligned.
        eval_points: Discretisation points of the functions.
        penalty: Penalisation term.
        grid_dim: Dimension of the grid used in the alignment algorithm.

    Returns:
        Array with the same shape than q_data with the srsf of
        the functions aligned to the template(s).

    """
    return optimum_reparam(
        np.ascontiguousarray(template_data.T),
        np.ascontiguousarray(eval_points),
        np.ascontiguousarray(q_data.T),
        method="DP2",
        lam=penalty,
        grid_dim=grid_dim,
    ).T


def _fisher_rao_warping_mean(
    warping: FDataGrid,
    *,
    max_iter: int = 100,
    tol: float = 1e-6,
    step_size: float = 0.3,
) -> FDataGrid:
    r"""
    Compute the karcher mean of a set of warpings.

    Let :math:`\gamma_i i=1...n` be a set of warping functions
    :math:`\gamma_i:[a,b] \rightarrow [a,b]` in :math:`\Gamma`, i.e.,
    monotone increasing and with the restriction :math:`\gamma_i(a)=a \,
    \gamma_i(b)=b`.

    The karcher mean :math:`\bar \gamma` is defined as the warping that
    minimises locally the sum of Fisher-Rao squared distances
    :footcite:`srivastava+klassen_2016_analysis_orbit`.

    .. math::
        \bar \gamma = argmin_{\gamma \in \Gamma} \sum_{i=1}^{n}
         d_{FR}^2(\gamma, \gamma_i)

    The computation is performed using the structure of Hilbert Sphere obtained
    after a transformation of the warpings, see
    :footcite:`srivastava++_2011_ficher-rao_orbit`.

    Args:
        warping: Set of warpings.
        max_iter: Maximum number of interations. Defaults to 100.
        tol: Convergence criterion, if the norm of the mean of the
            shooting vectors, :math:`| \bar v |<tol`, the algorithm will stop.
        step_size: Step size :math:`\epsilon` used to update the mean.

    Returns:
        Fdatagrid with the mean of the warpings. If
        shooting is True the shooting vectors will be returned in a tuple with
        the mean.

    References:
        .. footbibliography::

    """
    eval_points = warping.grid_points[0]
    original_eval_points = eval_points

    # Rescale warping to (0, 1)
    if warping.grid_points[0][0] != 0 or warping.grid_points[0][-1] != 1:

        eval_points = normalize_scale(eval_points)
        warping = FDataGrid(
            normalize_scale(warping.data_matrix[..., 0]),
            normalize_scale(warping.grid_points[0]),
        )

    # Compute srsf of warpings and their mean
    srsf = SRSF(output_points=eval_points, initial_value=0)
    psi = srsf.fit_transform(warping)

    # Find psi closest to the mean
    psi_centered = psi - srsf.fit_transform(warping.mean())
    psi_centered_data = psi_centered.data_matrix[..., 0]
    np.square(psi_centered_data, out=psi_centered_data)
    d = psi_centered_data.sum(axis=1).argmin()

    # Get raw values to calculate
    mu = np.atleast_2d(psi[d].data_matrix[0, ..., 0])
    psi_data = psi.data_matrix[..., 0]
    vmean = np.empty((1, len(eval_points)))

    # Construction of shooting vectors
    for _ in range(max_iter):

        vmean[0] = 0
        # Compute shooting vectors
        for psi_i in psi_data:

            inner = scipy.integrate.simps(mu * psi_i, x=eval_points)
            inner = max(min(inner, 1), -1)

            theta = np.arccos(inner)

            if theta > 1e-10:
                vmean += theta / np.sin(theta) * (psi_i - np.cos(theta) * mu)

        # Mean of shooting vectors
        vmean /= warping.n_samples
        v_norm = np.sqrt(scipy.integrate.simps(np.square(vmean)))

        # Convergence criterion
        if v_norm < tol:
            break

        # Calculate exponential map of mu
        a = np.cos(step_size * v_norm)
        b = np.sin(step_size * v_norm) / v_norm
        mu = a * mu + b * vmean

    # Recover mean in original gamma space
    warping_mean_ret = scipy.integrate.cumtrapz(
        np.square(mu, out=mu)[0],
        x=eval_points,
        initial=0,
    )

    # Affine traslation to original scale
    warping_mean_ret = normalize_scale(
        warping_mean_ret,
        a=original_eval_points[0],
        b=original_eval_points[-1],
    )

    monotone_interpolation = SplineInterpolation(
        interpolation_order=3,
        monotone=True,
    )

    return FDataGrid(
        [warping_mean_ret],
        grid_points=original_eval_points,
        interpolation=monotone_interpolation,
    )


def fisher_rao_karcher_mean(
    fdatagrid: FDataGrid,
    *,
    penalty: float = 0,
    center: bool = True,
    max_iter: int = 20,
    tol: float = 1e-3,
    initial: Optional[float] = None,
    grid_dim: int = 7,
    **kwargs: Any,
) -> FDataGrid:
    r"""
    Compute the Karcher mean under the elastic metric.

    Calculates the Karcher mean of a set of functional samples in the amplitude
    space :math:`\mathcal{A}=\mathcal{F}/\Gamma`.

    Let :math:`q_i` the corresponding SRSF of the observation :math:`f_i`.
    The space :math:`\mathcal{A}` is defined using the equivalence classes
    :math:`[q_i]=\{ q_i \circ \gamma \| \gamma \in \Gamma \}`, where
    :math:`\Gamma` denotes the space of warping functions. The karcher mean
    in this space is defined as

    .. math::
        [\mu_q] = argmin_{[q] \in \mathcal{A}} \sum_{i=1}^n
        d_{\lambda}^2([q],[q_i])

    Once :math:`[\mu_q]` is obtained it is selected the element of the
    equivalence class which makes the mean of the warpings employed be the
    identity.

    See :footcite:`srivastava+klassen_2016_analysis_karcher` and
    :footcite:`srivastava++_2011_ficher-rao_karcher`.

    Args:
        fdatagrid: Set of functions to compute the
            mean.
        penalty: Penalisation term. Defaults to 0.
        center: If ``True`` it is computed the mean of the warpings and
            used to select a central mean. Defaults ``True``.
        max_iter: Maximum number of iterations. Defaults to 20.
        tol: Convergence criterion, the algorithm will stop if
            :math:`|mu_{(\nu)} - mu_{(\nu - 1)}|_2 / | mu_{(\nu-1)} |_2 < tol`.
        initial: Value of the mean at the starting point. By default
            takes the average of the initial points of the samples.
        grid_dim: Dimension of the grid used in the alignment
            algorithm. Defaults 7.
        kwargs: Named options to be pased to :func:`_fisher_rao_warping_mean`.

    Returns:
        FDatagrid with the mean of the functions.

    Raises:
        ValueError: If the object is multidimensional or the shape of the srsf
            do not match with the fdatagrid.

    References:
        .. footbibliography::

    """
    check_is_univariate(fdatagrid)

    srsf_transformer = SRSF(initial_value=0)
    fdatagrid_srsf = srsf_transformer.fit_transform(fdatagrid)
    eval_points = fdatagrid.grid_points[0]

    eval_points_normalized = normalize_scale(eval_points)
    y_scale = eval_points[-1] - eval_points[0]

    interpolation = SplineInterpolation(interpolation_order=3, monotone=True)

    # Discretisation points
    fdatagrid_normalized = FDataGrid(
        fdatagrid(eval_points) / y_scale,
        grid_points=eval_points_normalized,
    )

    srsf = fdatagrid_srsf(eval_points)[..., 0]

    # Initialize with function closest to the L2 mean with the L2 distance
    centered = (srsf.T - srsf.mean(axis=0, keepdims=True).T).T

    distances = scipy.integrate.simps(
        np.square(centered, out=centered),
        eval_points_normalized,
        axis=1,
    )

    # Initialization of iteration
    mu = srsf[np.argmin(distances)]
    mu_aux = np.empty(mu.shape)
    mu_1 = np.empty(mu.shape)

    # Main iteration
    for _ in range(max_iter):

        gammas_matrix = _elastic_alignment_array(
            mu,
            srsf,
            eval_points_normalized,
            penalty,
            grid_dim,
        )

        gammas = FDataGrid(
            gammas_matrix,
            grid_points=eval_points_normalized,
            interpolation=interpolation,
        )

        fdatagrid_normalized = fdatagrid_normalized.compose(gammas)
        srsf = srsf_transformer.transform(
            fdatagrid_normalized,
        ).data_matrix[..., 0]

        # Next iteration
        mu_1 = srsf.mean(axis=0, out=mu_1)

        # Convergence criterion
        mu_norm = np.sqrt(
            scipy.integrate.simps(
                np.square(mu, out=mu_aux),
                eval_points_normalized,
            ),
        )

        mu_diff = np.sqrt(
            scipy.integrate.simps(
                np.square(mu - mu_1, out=mu_aux),
                eval_points_normalized,
            ),
        )

        if mu_diff / mu_norm < tol:
            break

        mu = mu_1

    if initial is None:
        initial = fdatagrid.data_matrix[:, 0].mean()

    srsf_transformer.set_params(initial_value=initial)

    # Karcher mean orbit in space L2/Gamma
    karcher_mean = srsf_transformer.inverse_transform(
        fdatagrid.copy(
            data_matrix=[mu],
            grid_points=eval_points,
            sample_names=("Karcher mean",),
        ),
    )

    if center:
        # Gamma mean in Hilbert Sphere
        mean_normalized = _fisher_rao_warping_mean(gammas, **kwargs)

        gamma_mean = FDataGrid(
            normalize_scale(
                mean_normalized.data_matrix[..., 0],
                a=eval_points[0],
                b=eval_points[-1],
            ),
            grid_points=eval_points,
        )

        gamma_inverse = invert_warping(gamma_mean)

        karcher_mean = karcher_mean.compose(gamma_inverse)

    # Return center of the orbit
    return karcher_mean



"""Outliergram Module.

This module contains the methods used to plot shapes in order to detect
shape outliers in our dataset. In order to do this, we plot the
Modified Band Depth and Modified Epigraph Index, that will help us detect
these outliers. The motivation of the method is that it is easy to find
magnitude outliers, but there is a necessity of capturing this other type.
"""

from typing import Optional, Sequence, Union

import numpy as np
from matplotlib.artist import Artist
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from ... import FDataGrid
from ..outliers import OutliergramOutlierDetector
from ._baseplot import BasePlot


class Outliergram(BasePlot):
    """
    Outliergram method of visualization.

    Plots the :class:`Modified Band Depth 
    (MBD)<skfda.exploratory.depth.ModifiedBandDepth>` on the Y axis and the
    :func:`Modified Epigraph Index
    (MEI)<skfda.exploratory.stats.modified_epigraph_index>` on the X axis.
    These points will create the form of a parabola.
    The shape outliers will be the points that appear far from this curve.

    Args:
        fdata: functional data set that we want to examine.
        chart: figure over with the graphs are plotted or axis over
            where the graphs are plotted. If None and ax is also
            None, the figure is initialized.
        fig: figure over with the graphs are plotted in case ax is not
            specified. If None and ax is also None, the figure is
            initialized.
        axes: axis where the graphs are plotted. If None, see param fig.
        n_rows: designates the number of rows of the figure
                to plot the different dimensions of the image. Only specified
                if fig and ax are None.
        n_cols: designates the number of columns of the
                figure to plot the different dimensions of the image. Only
                specified if fig and ax are None.
    Attributes:
        mbd: result of the calculation of the Modified Band Depth on our
            dataset. Represents the mean time a curve stays between other pair
            of curves, being a good measure of centrality.
        mei: result of the calculation of the Modified Epigraph Index on our
            dataset. Represents the mean time a curve stays below other curve.
    References:
        López-Pintado S.,  Romo J.. (2011). A half-region depth for functional
        data, Computational Statistics & Data Analysis, volume 55
        (page 1679-1695).
        Arribas-Gil A., Romo J.. Shape outlier detection and visualization for
        functional data: the outliergram
        https://academic.oup.com/biostatistics/article/15/4/603/266279
    """

    def __init__(
        self,
        fdata: FDataGrid,
        chart: Union[Figure, Axes, None] = None,
        *,
        fig: Optional[Figure] = None,
        axes: Optional[Axes] = None,
        factor: float = 1.5,
    ) -> None:
        BasePlot.__init__(
            self,
            chart,
            fig=fig,
            axes=axes,
        )
        self.fdata = fdata
        self.factor = factor
        self.outlier_detector = OutliergramOutlierDetector(factor=factor)
        self.outlier_detector.fit(fdata)
        indices = np.argsort(self.outlier_detector.mei_)
        self._parabola_ordered = self.outlier_detector.parabola_[indices]
        self._mei_ordered = self.outlier_detector.mei_[indices]

    @property
    def n_samples(self) -> int:
        return self.fdata.n_samples

    def _plot(
        self,
        fig: Figure,
        axes: Axes,
    ) -> None:

        self.artists = np.zeros(
            (self.n_samples, 1),
            dtype=Artist,
        )

        for i, (mei, mbd) in enumerate(
            zip(self.outlier_detector.mei_, self.outlier_detector.mbd_),
        ):
            self.artists[i, 0] = axes[0].scatter(
                mei,
                mbd,
                picker=2,
            )

        axes[0].plot(
            self._mei_ordered,
            self._parabola_ordered,
        )

        shifted_parabola = (
            self._parabola_ordered
            - self.outlier_detector.max_inlier_distance_
        )

        axes[0].plot(
            self._mei_ordered,
            shifted_parabola,
            linestyle='dashed',
        )

        # Set labels of graph
        if self.fdata.dataset_name is not None:
            axes[0].set_title(self.fdata.dataset_name)

        axes[0].set_xlabel("MEI")
        axes[0].set_ylabel("MBD")
        axes[0].set_xlim([0, 1])
        axes[0].set_ylim([
            0,  # Minimum MBD
            1,  # Maximum MBD
        ])



"""Test to check the Fda Feature Union module."""

import unittest

from pandas import DataFrame
from pandas.testing import assert_frame_equal

from skfda.datasets import fetch_growth
from skfda.misc.operators import SRSF
from skfda.preprocessing.dim_reduction.feature_extraction import (
    FDAFeatureUnion,
)
from skfda.preprocessing.smoothing.kernel_smoothers import (
    NadarayaWatsonSmoother,
)
from skfda.representation import EvaluationTransformer


class TestFDAFeatureUnion(unittest.TestCase):
    """Check the Fda Feature Union module."""

    def setUp(self) -> None:
        """Fetch the Berkeley Growth Study dataset."""
        self.X = fetch_growth(return_X_y=True)[0]

    def test_incompatible_fdatagrid_output(self) -> None:
        """Check that the transformer returns a fdatagrid."""
        u = FDAFeatureUnion(
            [("eval", EvaluationTransformer(None)), ("srsf", SRSF())],
            array_output=True,
        )
        self.assertRaises(TypeError, u.fit_transform, self.X)

    def test_correct_transformation_concat(self) -> None:
        """Check that the transformation is done correctly."""
        u = FDAFeatureUnion(
            [
                ("srsf1", SRSF()),
                ("smooth", NadarayaWatsonSmoother()),  # type: ignore
            ],
        )
        created_frame = u.fit_transform(self.X)

        t1 = SRSF().fit_transform(self.X)
        t2 = NadarayaWatsonSmoother().fit_transform(self.X)  # type: ignore

        true_frame = DataFrame({"Transformed data": [t1, t2]})
        assert_frame_equal(true_frame, created_frame)


if __name__ == '__main__':
    unittest.main()



"""Tests for elastic registration and functions in the SRVF framework."""

import unittest

import numpy as np

from skfda import FDataGrid
from skfda._utils import invert_warping, normalize_warping
from skfda.datasets import make_multimodal_samples, make_random_warping
from skfda.exploratory.stats import (
    _fisher_rao_warping_mean,
    fisher_rao_karcher_mean,
)
from skfda.misc.metrics import (
    PairwiseMetric,
    _fisher_rao_warping_distance,
    fisher_rao_amplitude_distance,
    fisher_rao_distance,
    fisher_rao_phase_distance,
    l2_distance,
)
from skfda.misc.operators import SRSF
from skfda.preprocessing.registration import FisherRaoElasticRegistration

metric = PairwiseMetric(l2_distance)
pairwise_fisher_rao = PairwiseMetric(fisher_rao_distance)


class TestFisherRaoElasticRegistration(unittest.TestCase):
    """Test elastic registration."""

    def setUp(self) -> None:
        """Initialize the samples."""
        template = make_multimodal_samples(n_samples=1, std=0, random_state=1)
        self.template = template
        self.template_rep = template.concatenate(
            template,
        ).concatenate(template)
        self.unimodal_samples = make_multimodal_samples(
            n_samples=3,
            random_state=1,
        )

        t = np.linspace(-3, 3, 9)
        self.dummy_sample = FDataGrid([np.sin(t)], t)

    def test_fit_wrong_dimension(self) -> None:
        """Checks that template and fit data is compatible."""
        reg = FisherRaoElasticRegistration(template=self.template)

        unimodal_samples = make_multimodal_samples(
            n_samples=3,
            points_per_dim=10,
            random_state=1,
        )

        with self.assertRaises(ValueError):
            reg.fit(unimodal_samples)

    def test_transform_wrong_dimension(self) -> None:
        """Checks that template and transform data is compatible."""
        reg = FisherRaoElasticRegistration(template=self.template)

        unimodal_samples = make_multimodal_samples(
            n_samples=3,
            points_per_dim=10,
            random_state=1,
        )

        reg.fit(self.unimodal_samples)
        with self.assertRaises(ValueError):
            reg.transform(unimodal_samples)

    def test_to_srsf(self) -> None:
        """Test to srsf."""
        # Checks SRSF conversion
        srsf = SRSF().fit_transform(self.dummy_sample)

        data_matrix = [
            [  # noqa: WPS317
                [-1.061897], [-0.75559027], [0.25355399],
                [0.81547327], [0.95333713], [0.81547327],
                [0.25355399], [-0.75559027], [-1.06189697],
            ],
        ]

        np.testing.assert_almost_equal(data_matrix, srsf.data_matrix)

    def test_from_srsf(self) -> None:
        """Test from srsf."""
        # Checks SRSF conversion
        srsf = SRSF(initial_value=0).inverse_transform(self.dummy_sample)

        data_matrix = [
            [  # noqa: WPS317
                [0], [-0.23449228], [-0.83464009],
                [-1.38200046], [-1.55623723], [-1.38200046],
                [-0.83464009], [-0.23449228], [0],
            ],
        ]

        np.testing.assert_almost_equal(data_matrix, srsf.data_matrix)

    def test_from_srsf_with_output_points(self) -> None:
        """Test from srsf."""
        # Checks SRSF conversion
        srsf_transformer = SRSF(
            initial_value=0,
            output_points=self.dummy_sample.grid_points[0],
        )
        srsf = srsf_transformer.inverse_transform(self.dummy_sample)

        data_matrix = [
            [  # noqa: WPS317
                [0], [-0.23449228], [-0.83464009],
                [-1.38200046], [-1.55623723], [-1.38200046],
                [-0.83464009], [-0.23449228], [0],
            ],
        ]

        np.testing.assert_almost_equal(data_matrix, srsf.data_matrix)

    def test_srsf_conversion(self) -> None:
        """Converts to srsf and pull backs."""
        srsf = SRSF()

        converted = srsf.fit_transform(self.unimodal_samples)
        converted = srsf.inverse_transform(converted)

        # Distances between original samples and s -> to_srsf -> from_srsf
        distances = np.diag(metric(converted, self.unimodal_samples))

        np.testing.assert_allclose(distances, 0, atol=8e-3)

    def test_template_alignment(self) -> None:
        """Test alignment to 1 template."""
        reg = FisherRaoElasticRegistration(template=self.template)
        register = reg.fit_transform(self.unimodal_samples)
        distances = metric(self.template, register)

        np.testing.assert_allclose(distances, 0, atol=12e-3)

    def test_one_to_one_alignment(self) -> None:
        """Test alignment to 1 sample to a template."""
        reg = FisherRaoElasticRegistration(template=self.template)
        register = reg.fit_transform(self.unimodal_samples[0])
        distances = metric(self.template, register)

        np.testing.assert_allclose(distances, 0, atol=12e-3)

    def test_set_alignment(self) -> None:
        """Test alignment 3 curves to set with 3 templates."""
        # Should give same result than test_template_alignment
        reg = FisherRaoElasticRegistration(template=self.template_rep)
        register = reg.fit_transform(self.unimodal_samples)
        distances = metric(self.template, register)

        np.testing.assert_allclose(distances, 0, atol=12e-3)

    def test_default_alignment(self) -> None:
        """Test alignment by default."""
        # Should give same result than test_template_alignment
        reg = FisherRaoElasticRegistration()
        register = reg.fit_transform(self.unimodal_samples)

        values = register([-0.25, -0.1, 0, 0.1, 0.25])

        expected = [
            [
                [0.599058], [0.997427], [0.772248], [0.412342], [0.064725],
            ],
            [
                [0.626875], [0.997155], [0.791649], [0.382181], [0.050098],
            ],
            [
                [0.620992], [0.997369], [0.785886], [0.376556], [0.048804],
            ],
        ]

        np.testing.assert_allclose(values, expected, atol=1e-4)

    def test_callable_alignment(self) -> None:
        """Test alignment by default."""
        # Should give same result than test_template_alignment
        reg = FisherRaoElasticRegistration(template=fisher_rao_karcher_mean)
        register = reg.fit_transform(self.unimodal_samples)

        values = register([-0.25, -0.1, 0, 0.1, 0.25])
        expected = [
            [
                [0.599058], [0.997427], [0.772248], [0.412342], [0.064725],
            ],
            [
                [0.626875], [0.997155], [0.791649], [0.382181], [0.050098],
            ],
            [
                [0.620992], [0.997369], [0.785886], [0.376556], [0.048804],
            ],
        ]

        np.testing.assert_allclose(values, expected, atol=1e-4)

    def test_simmetry_of_aligment(self) -> None:
        """Check registration using inverse composition."""
        reg = FisherRaoElasticRegistration(template=self.template)
        reg.fit_transform(self.unimodal_samples)
        warping = reg.warping_
        inverse = invert_warping(warping)
        register = self.template_rep.compose(inverse)
        distances = np.diag(metric(self.unimodal_samples, register))

        np.testing.assert_allclose(distances, 0, atol=12e-3)

    def test_raises(self) -> None:
        """Test that the assertions raise when appropriate."""
        reg = FisherRaoElasticRegistration()

        # Inverse transform without previous transform
        with np.testing.assert_raises(ValueError):
            reg.inverse_transform(self.unimodal_samples)

        # Inverse transform with different number of samples than transform
        reg.fit_transform(self.unimodal_samples)
        with np.testing.assert_raises(ValueError):
            reg.inverse_transform(self.unimodal_samples[0])

        # FDataGrid as template with n != 1 and n!= n_samples to transform
        reg = FisherRaoElasticRegistration(template=self.unimodal_samples).fit(
            self.unimodal_samples[0],
        )
        with np.testing.assert_raises(ValueError):
            reg.transform(self.unimodal_samples[0])

    def test_score(self) -> None:
        """Test score method of the transformer."""
        reg = FisherRaoElasticRegistration()
        reg.fit(self.unimodal_samples)
        score = reg.score(self.unimodal_samples)
        np.testing.assert_almost_equal(score, 0.999389)

    def test_warping_mean(self) -> None:
        """Test the warping_mean function."""
        warping = make_random_warping(start=-1, random_state=0)
        mean = _fisher_rao_warping_mean(warping)
        values = mean([-1, -0.5, 0, 0.5, 1])
        expected = [[[-1], [-0.376241], [0.136193], [0.599291], [1]]]
        np.testing.assert_array_almost_equal(values, expected)

    def test_linear(self) -> None:
        grid_points = [i for i in range(10)]
        data_matrix = np.array([grid_points, grid_points])
        fd = FDataGrid(
            data_matrix=data_matrix,
            grid_points=grid_points,
        )
        elastic_registration = FisherRaoElasticRegistration()
        fd_registered = elastic_registration.fit_transform(fd)
        np.testing.assert_array_almost_equal(
            fd_registered.data_matrix[..., 0],
            data_matrix,
        )


class TestElasticDistances(unittest.TestCase):
    """Test elastic distances."""

    def test_fisher_rao(self) -> None:
        """Test fisher rao distance."""
        t = np.linspace(0, 1, 100)
        sample = FDataGrid([t, 1 - t], t)
        f = np.square(sample)
        g = np.power(sample, 0.5)

        distance = [[0.64, 1.984], [1.984, 0.64]]
        res = pairwise_fisher_rao(f, g)

        np.testing.assert_almost_equal(res, distance, decimal=3)

    def test_fisher_rao_invariance(self) -> None:
        """Test invariance of fisher rao metric: d(f,g)= d(foh, goh)."""
        t = np.linspace(0, np.pi, 1000)
        id = FDataGrid([t], t)
        cos = np.cos(id)
        sin = np.sin(id)
        gamma = normalize_warping(np.sqrt(id), (0, np.pi))
        gamma2 = normalize_warping(np.square(id), (0, np.pi))

        distance_original = fisher_rao_distance(cos, sin)

        # Construction of 2 warpings
        distance_warping = fisher_rao_distance(
            cos.compose(gamma),
            sin.compose(gamma),
        )
        distance_warping2 = fisher_rao_distance(
            cos.compose(gamma2),
            sin.compose(gamma2),
        )

        # The error ~0.001 due to the derivation
        np.testing.assert_allclose(
            distance_original,
            distance_warping,
            atol=0.01,
        )

        np.testing.assert_allclose(
            distance_original,
            distance_warping2,
            atol=0.01,
        )

    def test_fisher_rao_amplitude_distance_limit(self) -> None:
        """Test limit of amplitude distance penalty."""
        f = make_multimodal_samples(n_samples=1, random_state=1)
        g = make_multimodal_samples(n_samples=1, random_state=9999)

        amplitude_limit = fisher_rao_amplitude_distance(f, g, lam=1000)
        fr_distance = fisher_rao_distance(f, g)

        np.testing.assert_almost_equal(amplitude_limit, fr_distance)

    def test_fisher_rao_phase_distance_id(self) -> None:
        """Test of phase distance invariance."""
        f = make_multimodal_samples(n_samples=1, random_state=1)

        phase = fisher_rao_phase_distance(f, 2 * f)

        np.testing.assert_allclose(phase, 0, atol=1e-7)

    def test_fisher_rao_warping_distance(self) -> None:
        """Test of warping distance."""
        t = np.linspace(0, 1, 1000)
        w1 = FDataGrid([t**5], t)
        w2 = FDataGrid([t**3], t)

        d = _fisher_rao_warping_distance(w1, w2)
        np.testing.assert_allclose(d, np.arccos(np.sqrt(15) / 4), atol=1e-3)

        d = _fisher_rao_warping_distance(w2, w2)
        np.testing.assert_allclose(d, 0, atol=2e-2)


if __name__ == '__main__':
    print()
    unittest.main()



from . import basis
from . import extrapolation
from . import grid
from . import interpolation
from ._evaluation_trasformer import EvaluationTransformer
from ._functional_data import FData
from .basis import FDataBasis
from .grid import FDataGrid



"""Tests of compatibility between numpy ufuncs and FDataGrid."""

import unittest
from typing import Any, Callable, TypeVar

import numpy as np
import pytest

import skfda
from skfda import FDataGrid


@pytest.fixture(params=[
    np.sqrt,
    np.absolute,
    np.round,
    np.exp,
    np.log,
    np.log10,
    np.log2,
])
def monary(request: Any) -> Any:
    """
    Fixture providing the monary function to validate.

    Not all of them are ufuncs.

    """
    return request.param


T = TypeVar("T", np.ndarray, FDataGrid)


def test_monary_ufuncs(monary: Callable[[T], T]) -> None:
    """Test that unary ufuncs can be applied to FDataGrid."""
    data_matrix = np.arange(15).reshape(3, 5) + 1

    fd = FDataGrid(data_matrix)

    fd_monary = monary(fd)

    fd_monary_build = FDataGrid(monary(data_matrix))

    assert fd_monary.equals(fd_monary_build)


def test_binary_ufunc() -> None:
    """Test that binary ufuncs can be applied to FDataGrid."""
    data_matrix = np.arange(15).reshape(3, 5)
    data_matrix2 = 2 * np.arange(15).reshape(3, 5)

    fd = FDataGrid(data_matrix)
    fd2 = FDataGrid(data_matrix2)

    fd_mul = np.multiply(fd, fd2)

    fd_mul_build = FDataGrid(data_matrix * data_matrix2)

    assert fd_mul.equals(fd_mul_build)


def test_out_ufunc(monary: Callable[..., Any]) -> None:
    """Test that the out parameter of ufuncs work for FDataGrid."""
    data_matrix = np.arange(15).reshape(3, 5) + 1
    data_matrix_copy = np.copy(data_matrix)

    fd = FDataGrid(data_matrix)

    monary(fd, out=fd)

    fd_monary_build = FDataGrid(monary(data_matrix_copy))

    assert fd.equals(fd_monary_build)


class TestOperators(unittest.TestCase):
    """Tests for operators."""

    def test_commutativity(self) -> None:
        """Test that operations with numpy arrays commute."""
        X = FDataGrid([[1, 2, 3], [4, 5, 6]])
        arr = np.array([1, 2])

        self.assertTrue((arr + X).equals((X + arr)))

    def test_commutativity_basis(self) -> None:
        """Test that operations with numpy arrays for basis commute."""
        X = FDataGrid([[1, 2, 3], [4, 5, 6]])
        arr = np.array([1, 2])
        basis = skfda.representation.basis.Fourier(n_basis=5)

        X_basis = X.to_basis(basis)

        self.assertTrue((arr * X_basis).equals((X_basis * arr)))
